/*************************************************************************
|  COPYRIGHT (c) 2000 BY ABATRON AG
|*************************************************************************
|
|  PROJECT NAME: BDI Interface Library
|  FILENAME    : bdiifc.c
|
|  COMPILER    : GCC
|
|  TARGET OS   : LINUX
|  TARGET HW   : PC
|
|  PROGRAMMER  : Abatron / RD
|  CREATION    : 22.06.2000
|
|*************************************************************************
|
|  DESCRIPTION :
|  BDI access functions.
|
|*************************************************************************/

/*************************************************************************
|  INCLUDES
|*************************************************************************/

#include <unistd.h>
#include <sys/times.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>

#include "bdierror.h"
#include "bdicmd.h"
#include "bdilnk.h"
#include "bdiifc.h"

/*************************************************************************
|  DEFINES
|*************************************************************************/

#define BDI_DEFAULT_EXEC_TIME       100 /* default BDI command execution time */

/* target error codes */
#define TARGET_ERR_STATE          -2101
#define TARGET_ERR_PARAMETER      -2102
#define TARGET_ERR_STEP_TIMEOUT   -2103
#define TARGET_ERR_RESOURCE       -2104
#define TARGET_ERR_VERIFY         -2105
#define TARGET_ERR_WRITE          -2106

/* Write Modes */
#define BDI_LOAD_MODE_RAM       0   /* Write into target RAM */
#define BDI_LOAD_MODE_VERIFY    1   /* Verify target memory  */

/* HC12 memory Type */
#define HC12_MEM_INT_EEPROM     0
#define HC12_MEM_INT_FLASH      1
#define HC12_MEM_EXTERNAL       2

/* common bit constants */
#define BIT0      (1L<<0)
#define BIT1      (1L<<1)
#define BIT2      (1L<<2)
#define BIT3      (1L<<3)
#define BIT4      (1L<<4)
#define BIT5      (1L<<5)
#define BIT6      (1L<<6)
#define BIT7      (1L<<7)
#define BIT8      (1L<<8)
#define BIT9      (1L<<9)
#define BIT10     (1L<<10)
#define BIT11     (1L<<11)
#define BIT12     (1L<<12)
#define BIT13     (1L<<13)
#define BIT14     (1L<<14)
#define BIT15     (1L<<15)

#define BIT16     (1L<<16)
#define BIT17     (1L<<17)
#define BIT18     (1L<<18)
#define BIT19     (1L<<19)
#define BIT20     (1L<<20)
#define BIT21     (1L<<21)
#define BIT22     (1L<<22)
#define BIT23     (1L<<23)
#define BIT24     (1L<<24)
#define BIT25     (1L<<25)
#define BIT26     (1L<<26)
#define BIT27     (1L<<27)
#define BIT28     (1L<<28)
#define BIT29     (1L<<29)
#define BIT30     (1L<<30)
#define BIT31     (1L<<31)


/*************************************************************************
|  MACROS
|*************************************************************************/

/*************************************************************************
|  TYPEDEFS
|*************************************************************************/

/*************************************************************************
|  LOCALS
|*************************************************************************/

static BYTE     cmdBuffer[BDI_MAX_FRAME_SIZE];
static BYTE     ansBuffer[BDI_MAX_FRAME_SIZE];

static BYTE     memWriteMode;   /* Current Write Mode (0=RAM,1=verify,2..n=flash    */
static BYTE     memFlashType;   /* Flash Type (write mode)                          */
static BYTE     memBusWidth;    /* Target memory bus width (1,2,4 bytes)            */
static BYTE     memFlashWidth;  /* Flash chip width (1,2,4 bytes)                   */
static DWORD    memPageSize;    /* the size of a page in target memory              */
static DWORD    memBankMask;    /* Mask used to get the base of the memory bank     */
static DWORD    memRamBase;     /* Base address of RAM in the target system for     */
                                /* the algorithm code                               */
static DWORD    memOption1;     /* an optional parameter, depends on write mode     */
static DWORD    memFlashSize;   /* size of the flash memory                         */
static BOOL     flashChanged;   /* set after a call to BDI_FlashSetType() */

static BOOL     targetIsHC12;   /* set if target is HC12 */
static BOOL     targetIsS12X;   /* set if target is S12X */
static BYTE     memTypeHC12;    /* HC12 memory type */
static WORD     cpuTypeHC12;    /* the HC12 CPU type */
static BOOL     pagedAddress;   /* set if file has paged addresses */

static BDI_ProgressCallbackT* progressCallback = NULL;

#undef _DEBUG
#ifdef _DEBUG
static FILE*    logFile = NULL; /* logging file   */
#endif


/****************************************************************************
 ****************************************************************************/

void flush_logFile(void)
 {
  #ifdef _DEBUG
      printf ("    Flushing debug data to file...\n");
      if (logFile != NULL) fflush(logFile);
  #endif
} /* flush_logFile */

/****************************************************************************
 ****************************************************************************

    BDI_Append___ :

    Host independent helper function to append a numeric value to a buffer.
    The bytes will be stored in the link format order (Motorola byte order).

     INPUT  : value     value
              buffer    pointer to buffer
     OUTPUT : RETURN    pointer to next byte after the stored value

 ****************************************************************************/

static BYTE* BDI_AppendChar(char value, BYTE* buffer)
{
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendCHAR */

static BYTE* BDI_AppendByte(BYTE  value, BYTE* buffer)
{
  *buffer++ = value;
  return buffer;
} /* BDI_AppendByte */

static BYTE* BDI_AppendWord(WORD value, BYTE* buffer)
{
  *buffer++ = (BYTE)(value>>8);
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendWord */

static BYTE* BDI_AppendLong(DWORD value, BYTE* buffer)
{
  *buffer++ = (BYTE)(value>>24);
  *buffer++ = (BYTE)(value>>16);
  *buffer++ = (BYTE)(value>>8);
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendLong */


/****************************************************************************
 ****************************************************************************

    BDI_Extract___ :

    Host independent helper function to extract a numeric value from a buffer.
    The value must be stored in link format (Motorola byte order).

     INPUT  : buffer    pointer to buffer
     OUTPUT : value     pointer to variable where to store the value
              RETURN    pointer to next byte after the extracted value

 ****************************************************************************/

static BYTE* BDI_ExtractChar(char* value, BYTE* buffer)
{
  *value = (char)*buffer++;
  return buffer;
} /* BDI_ExtractCHAR */

static BYTE* BDI_ExtractByte(BYTE* value, BYTE* buffer)
{
  *value = *buffer++;
  return buffer;
} /* BDI_ExtractByte */

static BYTE* BDI_ExtractWord(WORD* value, BYTE* buffer)
{
  WORD x;
  x = (WORD)*buffer++;
  x = (x<<8) + (WORD)*buffer++;
  *value = x;
  return buffer;
} /* BDI_ExtractWord */

static BYTE* BDI_ExtractLong(DWORD* value, BYTE* buffer)
{
  DWORD x;
  x = (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  *value = x;
  return buffer;
} /* BDI_ExtractLong */


/****************************************************************************
 ****************************************************************************
    BDI_DoDelay
    Helper function to delay an amount of time

     INPUT:  delay        the delay time in ms
     OUTPUT: -

 ****************************************************************************/

static void BDI_DoDelay(DWORD delay)
{
  usleep(1000 * delay);
} /* BDI_DoDelay */


/****************************************************************************
 ****************************************************************************
    AppendHex

    Append a byte to the S-record

 ****************************************************************************/

static char* AppendHex(char* szRecord, BYTE value, WORD* checkSum)
{
  static char ByteToHex[16] = {'0','1','2','3','4','5','6','7','8','9',
                               'A','B','C','D','E','F'};
  *checkSum  += value;
  *szRecord++ = ByteToHex[value >> 4];
  *szRecord++ = ByteToHex[value & 0x0F];
  return szRecord;
}


/****************************************************************************
 ****************************************************************************
    BuildSRecord

    Build a S-Record string.

 ****************************************************************************/

static size_t BuildSRecord(      char    chType, /* the type */
                                 DWORD   dwAddr, /* the address */
                                 WORD    nCount, /* the number of bytes in the data */
                           const BYTE*   data,   /* the data part */
                                 char*   szLine) /* the S-Record buffer */
{
  WORD      addrLen;
  WORD      checkSum;
  char*     putPtr;

  switch (chType)
  {
    case '0':
    case '1':
    case '9':
      addrLen = 2;
      break;
    case '2':
    case '8':
      addrLen = 3;
      break;
    case '3':
    case '7':
      addrLen = 4;
      break;
    default:
      return 0;
  } /* end switch */

  /* write start */
  putPtr    = szLine;
  *putPtr++ = 'S';
  *putPtr++ = chType;

  /* append LEN field */
  checkSum = 0;
  putPtr = AppendHex(putPtr, (BYTE)(addrLen + nCount + 1), &checkSum);


  /* append address field */
  if (addrLen == 4) putPtr = AppendHex(putPtr, (BYTE)(dwAddr >> 24), &checkSum);
  if (addrLen >= 3) putPtr = AppendHex(putPtr, (BYTE)(dwAddr >> 16), &checkSum);
  putPtr = AppendHex(putPtr, (BYTE)(dwAddr >> 8), &checkSum);
  putPtr = AppendHex(putPtr, (BYTE)(dwAddr >> 0), &checkSum);

  /* append code/data fields */
  while (nCount--) putPtr = AppendHex(putPtr, *data++, &checkSum);

  /* append checksum field */
  putPtr = AppendHex(putPtr, (BYTE)(~checkSum), &checkSum);

  /* append new line and terminate string*/
  *putPtr++ = '\n';
  *putPtr   = 0;

  return putPtr - szLine;
}


/****************************************************************************
 ****************************************************************************

    DecodeSRecord:

    Decode an Data S-Record (S1,S2,S3).

    INPUT  : sRecord      the S-Record to decode
    OUTPUT : addrPtr      the address for the decoded data
             dataPtr      the data part of the record (binary)
             RETURN       number of databytes, 0 if no data record, -1 if error

 ****************************************************************************/

#define SREC_MAX_LINE   520   /* 4 + (2 x 255) + 2 + 1 = 517 */
#define SREC_MAX_DATA   256

static BYTE Hex2Bin(char highDigit, char lowDigit)
{
  static BYTE HexTableLo[] = {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,
                              0,0,0,0,0,0,0,         /* :;<=>?@ */
                              0x0A,0x0B,0x0C,0x0D,0x0E,0x0F};

  static BYTE HexTableHi[] = {0x00,0x10,0x20,0x30,0x40,0x50,0x60,0x70,0x80,0x90,
                              0,0,0,0,0,0,0,         /* :;<=>?@ */
                              0xA0,0xB0,0xC0,0xD0,0xE0,0xF0};

  return  (BYTE)(   (HexTableHi[(highDigit - '0') & 0x1F])
                  + (HexTableLo[(lowDigit  - '0') & 0x1F]) );
} /* Hex2Bin */


static int DecodeSRecord(char* sRecord, DWORD* addrPtr, BYTE* dataPtr)
{
  int       count;
  int       recLen;
  int       addrLen;
  int       i;
  char      recType;
  BYTE      checksum;
  BYTE      nextValue;
  DWORD     address;

  if (*sRecord <= ' ') return 0;  /* ignore empty line */
  if (*sRecord++ != 'S') return -1;
  recType   = *sRecord++;
  nextValue = Hex2Bin(*sRecord, *(sRecord+1));
  sRecord  += 2;
  checksum  = nextValue;
  recLen    = nextValue;

  /* extract address */
  if      (recType == '1') addrLen = 2;
  else if (recType == '2') addrLen = 3;
  else if (recType == '3') addrLen = 4;
  else                     return 0;
  address = 0;
  for (i=0; i<addrLen; i++) {
    nextValue = Hex2Bin(*sRecord, *(sRecord+1));
    sRecord  += 2;
    checksum  = (BYTE)(checksum + nextValue);
    address   = (address << 8) + nextValue;
  } /* for */
  *addrPtr = address;

  /* get data */
  count = recLen - addrLen - 1;
  for (i=0; i<count; i++) {
    nextValue = Hex2Bin(*sRecord, *(sRecord+1));
    sRecord  += 2;
    checksum  = (BYTE)(checksum + nextValue);
    *dataPtr++ = nextValue;
  } /* for */

  /* check sum */
  nextValue = Hex2Bin(*sRecord, *(sRecord+1));
  checksum = (BYTE)(checksum + nextValue);
  if (checksum != 0xFF) return -1;
  else                  return count;
} /* DecodeSRecord */


/****************************************************************************
 ****************************************************************************

    Set write mode for the next fast download

    INPUT:
    OUTPUT: return          error code

 ****************************************************************************/

static int BDI_SetWriteMode(BYTE   writeMode,
                            BYTE   busWidth,
                            DWORD  bankMask,
                            DWORD  ramBase,
                            DWORD  option1)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  BYTE      result;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_PRO_PROGRAM_SETUP, cmdBuffer);
  cmdPtr = BDI_AppendByte(writeMode, cmdPtr);
  cmdPtr = BDI_AppendByte(busWidth,  cmdPtr);
  cmdPtr = BDI_AppendLong(bankMask,  cmdPtr);
  cmdPtr = BDI_AppendLong(ramBase,   cmdPtr);
  cmdPtr = BDI_AppendLong(option1,   cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer, cmdBuffer, sizeof ansBuffer, ansBuffer, 3000);
  if (rxCount < 0) return rxCount;

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 2) || (answer != BDI_PRO_PROGRAM_SETUP))
    return BDI_ERR_INVALID_RESPONSE;

  ansPtr = BDI_ExtractByte(&result, ansPtr);
  if      (result == 0) return BDI_OKAY;
  else if (result == 1) return BDI_ERR_INVALID_MODE;
  else if (result == 3) return BDI_ERR_WORKING_RAM;
  else if (result == 4) return BDI_ERR_INVALID_PARAMETER;
  else                  return BDI_ERR_INVALID_RESPONSE;
} /* BDI_SetWriteMode */


/****************************************************************************
 ****************************************************************************

    BDI_Connect:

     Opens the connection to the BDI.

     INPUT  : host      a string for the comm port "/dev/..." or IP number
              param     the baudrate for serial connection
                        the port number for UDP/IP connection
     OUTPUT : RETURN    0 if okay or a negativ number if error.

 ****************************************************************************/


int BDI_Connect(const char *host, DWORD param)
{
  BYTE     *cmdPtr;
  int       rxCount;
  int       result;

#ifdef _DEBUG
  if (logFile == NULL)logFile = fopen("bdiifc.log","w");
  if (logFile != NULL) {
    fprintf(logFile, "BDI_Connect(%s, %lu)", host, param);
  } /* if */
#endif

  /* init locals */
  memWriteMode     = BDI_LOAD_MODE_RAM;
  memFlashType     = BDI_LOAD_MODE_RAM;
  memBusWidth      = 2;
  memFlashSize     = 0;
  memBankMask      = 0;
  memRamBase       = 0;
  memOption1       = 0;
  flashChanged     = TRUE;
  targetIsHC12     = FALSE;
  targetIsS12X     = FALSE;
  memTypeHC12      = HC12_MEM_INT_EEPROM;
  cpuTypeHC12      = BDI_CPU_HC812A4;
  pagedAddress     = FALSE;
  progressCallback = NULL;

  /* connect to BDI */
  result = BDI_Open(host, param);
  if (result < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(result));
    } /* if */
#endif
    return result;
  } /* if */

  /* send exit loader command */
  cmdPtr  = BDI_AppendByte(BDI_LDR_EXIT_LOADER, cmdBuffer);
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
    BDI_Close();
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } // if

  /* delay and connect again if firmware is not already activ,  */
  /* firmware answers with BDI_ANS_NOT_IMPLEMENTED              */
  if (*ansBuffer == BDI_LDR_EXIT_LOADER) {
    BDI_Close();
    BDI_DoDelay(2000);
    result = BDI_Open(host, param);
    if (result < 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(result));
      } /* if */
#endif
      return result;
    } /* if */
  } /* if */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return result;
} /* BDI_Connect */


/****************************************************************************
 ****************************************************************************

    BDI_Disconnect:

     Closes the connection to the BDI.

     INPUT  : -
     OUTPUT : RETURN      0 if okay or a negativ number if error.

 ****************************************************************************/

int BDI_Disconnect(void)
{
#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_Disconnect()\n");
  } /* if */
#endif
  return BDI_Close();
} /* BDI_Disconnect */


/****************************************************************************
 ****************************************************************************

    Read target registers

    INPUT:  count         the number of registers to read
            pRegList      points to the list with the registers to read
    OUTPUT: pValues       points to the array where to store the results
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_RegisterGet(      WORD          count,
                    const BDI_RegTypeT *pRegList,
                          DWORD        *pValues)
{
  BYTE      *cmdPtr;
  BYTE      *ansPtr;
  int        rxCount;
  BYTE       answer;
  WORD       errorCode;
  WORD       regCount;
  DWORD     *pPut;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_RegisterGet(%u, (%u, %u), ...)", count, pRegList->regType, pRegList->regAddr);
  } /* if */
#endif

  /* prepare command */
  regCount = count;
  cmdPtr = BDI_AppendByte(BDI_CMD_REGISTER_GET, cmdBuffer);
  cmdPtr = BDI_AppendWord(regCount, cmdPtr);
  while (regCount > 0) {
    cmdPtr = BDI_AppendWord(pRegList->regType, cmdPtr);
    cmdPtr = BDI_AppendWord(pRegList->regAddr, cmdPtr);
    regCount--;
    pRegList++;
  } /* while */

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if (answer != BDI_CMD_REGISTER_GET) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  if (errorCode == 0) {
    ansPtr = BDI_ExtractWord(&regCount, ansPtr);
    pPut   = pValues;
    while (regCount > 0) {
      ansPtr = BDI_ExtractLong(pPut, ansPtr);
      regCount--;
      pPut++;
    } /* while */
  } /* if */

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY (0x%08lx, ...)\n", *pValues);
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_RegisterGet */


/****************************************************************************
 ****************************************************************************

    Write to target registers

    INPUT:  count         the number of registers to write
            pRegList      points to the list with the registers to write
            pValues       points to the array with the values to write
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_RegisterSet(      WORD          count,
                    const BDI_RegTypeT *pRegList,
                    const DWORD        *pValues)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      regCount;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_RegisterSet(%u, (%u, %u, 0x%08lx), ...)", count, pRegList->regType, pRegList->regAddr, *pValues);
  } /* if */
#endif

  /* prepare command */
  regCount = count;
  cmdPtr = BDI_AppendByte(BDI_CMD_REGISTER_SET, cmdBuffer);
  cmdPtr = BDI_AppendWord(regCount, cmdPtr);
  while (regCount > 0) {
    cmdPtr = BDI_AppendWord(pRegList->regType, cmdPtr);
    cmdPtr = BDI_AppendWord(pRegList->regAddr, cmdPtr);
    cmdPtr = BDI_AppendLong(*pValues, cmdPtr);
    regCount--;
    pRegList++;
    pValues++;
  } /* while */

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_REGISTER_SET)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_RegisterSet */


/****************************************************************************
 ****************************************************************************

    Read from floating-point register(s)

    INPUT:  first         the first register to read
            last          the last register to read
    OUTPUT: data          the read data (8 bytes for every register)
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FPRegGet(WORD first, WORD last, BYTE* data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      dataCount;

#ifdef _DEBUG
  BYTE* val = data;
  if (logFile != NULL) {
    fprintf(logFile, "BDI_FPRegGet(%u, %u)", first, last);
  } /* if */
#endif

  /* check parameters */
  dataCount = last - first + 1;
  if (dataCount > 128) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */
  dataCount = 8 * dataCount;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_FPREG_GET, cmdBuffer);
  cmdPtr = BDI_AppendWord(first, cmdPtr);
  cmdPtr = BDI_AppendWord(last, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if (answer != BDI_CMD_FPREG_GET) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  if (errorCode == 0) {
    while (dataCount--) *data++ = *ansPtr++;
  } /* if */

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " 0x%08lx->%02x%02x%02x%02x%02x%02x%02x%02x",
                         val, val[0],val[1],val[2],val[3],val[4],val[5],val[6],val[7]);
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_FPRegGet */


/****************************************************************************
 ****************************************************************************

    Write to floating-point register

    INPUT:  first         the first register to read
            last          the last register to read
            data          data to write (8 bytes for every register)
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FPRegSet(WORD first, WORD last, const BYTE* data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      dataCount;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_FPRegSet(%u, %u, 0x%08lx->%02x%02x%02x%02x%02x%02x%02x%02x)",
            first,last,data,data[0],data[1],data[2],data[3],data[4],data[5],data[6],data[7]);
  } /* if */
#endif

  /* check parameters */
  dataCount = last - first + 1;
  if (dataCount > 128) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */
  dataCount = 8 * dataCount;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_FPREG_SET, cmdBuffer);
  cmdPtr = BDI_AppendWord(first, cmdPtr);
  cmdPtr = BDI_AppendWord(last, cmdPtr);
  while (dataCount--) *cmdPtr++ = *data++;

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_FPREG_SET)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_FPRegSet */


/****************************************************************************
 ****************************************************************************

    Read from vector register(s)

    INPUT:  first         the first register to read
            last          the last register to read
    OUTPUT: data          the read data (16 bytes for every register)
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_VRegGet(WORD first, WORD last, BYTE* data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      dataCount;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_VRegGet(%u, %u)", first, last);
  } /* if */
#endif

  /* check parameters */
  dataCount = last - first + 1;
  if (dataCount > 64) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */
  dataCount = 16 * dataCount;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_VREG_GET, cmdBuffer);
  cmdPtr = BDI_AppendWord(first, cmdPtr);
  cmdPtr = BDI_AppendWord(last, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if (answer != BDI_CMD_VREG_GET) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  if (errorCode == 0) {
    while (dataCount--) *data++ = *ansPtr++;
  } /* if */

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_VRegGet */


/****************************************************************************
 ****************************************************************************

    Write to vector register

    INPUT:  first         the first register to read
            last          the last register to read
            data          data to write (8 bytes for every register)
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_VRegSet(WORD first, WORD last, const BYTE* data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      dataCount;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_VRegSet(%u, %u)", first, last);
  } /* if */
#endif

  /* check parameters */
  dataCount = last - first + 1;
  if (dataCount > 64) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */
  dataCount = 16 * dataCount;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_VREG_SET, cmdBuffer);
  cmdPtr = BDI_AppendWord(first, cmdPtr);
  cmdPtr = BDI_AppendWord(last, cmdPtr);
  while (dataCount--) *cmdPtr++ = *data++;

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_VREG_SET)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_VRegSet */


/****************************************************************************
 ****************************************************************************

    Read from the target memory

    INPUT:  addr          address to read from
            space         the memory space to use
            count         number of bytes to read
    OUTPUT: data          the read data will be stored in this buffer
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_MemoryRead(DWORD addr, WORD space, WORD count, BYTE *data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  WORD      currentCount;
  BYTE      answer;
  WORD      errorCode;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_MemoryRead(0x%08lx, %u, %u)", addr, space, count);
  } /* if */
#endif

  while (count > 0) {

    /* get next partial block */
    if (count > BDI_MAX_BLOCK_SIZE) currentCount = BDI_MAX_BLOCK_SIZE;
    else                            currentCount = count;

    /* prepare command */
    cmdPtr = BDI_AppendByte(BDI_CMD_MEMORY_READ, cmdBuffer);
    cmdPtr = BDI_AppendLong(addr,         cmdPtr);
    cmdPtr = BDI_AppendWord(space,        cmdPtr);
    cmdPtr = BDI_AppendWord(currentCount, cmdPtr);
    count -= currentCount;
    addr  += currentCount;

    /* BDI transaction */
    rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                              cmdBuffer,
                              sizeof ansBuffer,
                              ansBuffer,
                              200 + currentCount);
    if (rxCount < 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
      } /* if */
#endif
      return rxCount;
    } /* if */

    /* analyse response */
    ansPtr = ansBuffer;
    answer = *ansPtr++;
    if ((rxCount != (int)(currentCount+3)) || (answer != BDI_CMD_MEMORY_READ)) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
      } /* if */
#endif
      return BDI_ERR_INVALID_RESPONSE;
    } /* if */

    ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
    while (currentCount--) *data++ = *ansPtr++;
    if (errorCode != 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
      } /* if */
#endif
      return (int)(short)errorCode;
    } /* if */
  } /* while */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_MemoryRead */


/****************************************************************************
 ****************************************************************************

    Write to the target memory

    INPUT:  addr          address to write to
            space         the memory space to use
            count         number of bytes to write
            data          data to write to the target memory
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_MemoryWrite(DWORD addr, WORD space, WORD count, const BYTE *data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  WORD      currentCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      i;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_MemoryWrite(0x%08lx, %u, %u)", addr, space, count);
  } /* if */
#endif

  while (count > 0) {

    /* get next partial block */
    if (count > BDI_MAX_BLOCK_SIZE) currentCount = BDI_MAX_BLOCK_SIZE;
    else                            currentCount = count;

    /* prepare command */
    cmdPtr = BDI_AppendByte(BDI_CMD_MEMORY_WRITE, cmdBuffer);
    cmdPtr = BDI_AppendLong(addr,         cmdPtr);
    cmdPtr = BDI_AppendWord(space,        cmdPtr);
    cmdPtr = BDI_AppendWord(currentCount, cmdPtr);
    for (i=0; i<currentCount; i++) *cmdPtr++ = *data++;
    count -= currentCount;
    addr  += currentCount;

    /* BDI transaction */
    rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                              cmdBuffer,
                              sizeof ansBuffer,
                              ansBuffer,
                              200 + currentCount);
    if (rxCount < 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
      } /* if */
#endif
      return rxCount;
    } /* if */

    /* analyse response */
    ansPtr = ansBuffer;
    answer = *ansPtr++;
    if ((rxCount != 3) || (answer != BDI_CMD_MEMORY_WRITE)) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
      } /* if */
#endif
      return BDI_ERR_INVALID_RESPONSE;
    } /* if */

    ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
    if (errorCode != 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
      } /* if */
#endif
      return (int)(short)errorCode;
    } /* if */
  } /* while */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_MemoryWrite */


/****************************************************************************
 ****************************************************************************

    Start the target CPU

    INPUT:  -
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_TargetStart(void)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_TargetStart()");
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_TARGET_START, cmdBuffer);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            3000);  /* because of long breakpoint inser time */
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_TARGET_START)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_TargetStart */


/****************************************************************************
 ****************************************************************************

    Stop execution of target code
    The command is repeated in order to use different Control bytes.
    This because the Control byte may equal to DLE and so the resulting
    RS232 frame is greater than the Z85230 Fifo size which is 8.
    This may be a problem with a HC12 target at very low BDM clock speed.

    INPUT:  -
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_TargetHalt(void)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      repeatCount;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_TargetHalt()");
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_TARGET_HALT, cmdBuffer);

  /* BDI transaction */
  repeatCount = 0;
  do {
    rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                              cmdBuffer,
                              sizeof ansBuffer,
                              ansBuffer,
                              3000);  /* because of long breakpoint inser time */
    repeatCount++;
  } while ((rxCount < 0) && (repeatCount < 3));

  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_TARGET_HALT)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_TargetHalt */


/****************************************************************************
 ****************************************************************************

    Reset the target.

    INPUT:  -
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

#define BDI_RESET_DEFAULT       0
#define BDI_RESET_FREEZE        1
#define BDI_RESET_RUN           2

int BDI_TargetReset2(WORD resetTime)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  DWORD     executionTime;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_TargetReset()");
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_RESET_TARGET, cmdBuffer);
  cmdPtr = BDI_AppendByte(BDI_RESET_DEFAULT, cmdPtr);
  cmdPtr = BDI_AppendWord(resetTime, cmdPtr);

  /* BDI transaction */
  if (resetTime > 0) executionTime = resetTime + 3000;
  else               executionTime = 5000;
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            executionTime);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 1) || (answer != BDI_CMD_RESET_TARGET)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  memWriteMode = BDI_LOAD_MODE_RAM; /* write mode has changed to RAM */
#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_TargetReset2 */


int BDI_TargetReset(void)
{
  return BDI_TargetReset2(0);
} /* BDI_TargetReset */


int BDI_TargetReset_mod(WORD resetTime)
{
  return BDI_TargetReset2(resetTime);
} /* BDI_TargetReset_mod */


/****************************************************************************
 ****************************************************************************

    Single step a range

    INPUT:  lowerBound    lower bound of single step range
            upperBound    upper bound of single step range
            stepCount     maximal number of steps to execute
    OUTPUT: pc            the current program counter
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_TargetStep(DWORD  lowerBound,
                   DWORD  upperBound,
                   WORD   stepCount,
                   DWORD *pc)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_TargetStep(0x%08lx, 0x%08lx, %u)", lowerBound, upperBound, stepCount);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_TARGET_STEP, cmdBuffer);
  cmdPtr = BDI_AppendLong(lowerBound, cmdPtr);
  cmdPtr = BDI_AppendLong(upperBound, cmdPtr);
  cmdPtr = BDI_AppendWord(stepCount,  cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME + stepCount);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 7) || (answer != BDI_CMD_TARGET_STEP)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  ansPtr = BDI_ExtractLong(pc, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY (pc=0x%08lx)\n", *pc);
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_TargetStep */


/****************************************************************************
 ****************************************************************************

    Add a new software breakpoint

    INPUT:  addr          the address where to set a breakpoint
            interval      Number of breaks before the target is stopped
                            0 = temporary breakpoint, automatically removed
                            1 = standard breakpoint
                            n = breakpoint must be reached n times
    OUTPUT: RETURN        break identification or a negativ number if error

 ****************************************************************************/

int BDI_BreakAdd(DWORD addr, DWORD interval)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      breakId;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_BreakAdd(0x%08lx, %lu)", addr, interval);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_BREAK_ADD, cmdBuffer);
  cmdPtr = BDI_AppendLong(addr, cmdPtr);
  cmdPtr = BDI_AppendLong(interval, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 5) || (answer != BDI_CMD_BREAK_ADD)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  if (errorCode == 0) {
    ansPtr = BDI_ExtractWord(&breakId, ansPtr);
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY (id=%i)\n", breakId);
    } /* if */
#endif
    return (int)breakId;
  } /* if */
  else {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
#endif
    return (int)(short)errorCode;
  } /* else */
} /* BDI_BreakAdd */


/****************************************************************************
 ****************************************************************************

    Read back the number of times the breakpoint must be reached until
    the target is stopped

    INPUT:  breakId       the ID returned by the BDI_BreakAdd() function
    OUTPUT: count         Number of breaks before the target will be stopped
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_BreakRead(int breakId, DWORD *count)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_BreakRead(%i)", breakId);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_BREAK_READ, cmdBuffer);
  cmdPtr = BDI_AppendWord((WORD)breakId, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 7) || (answer != BDI_CMD_BREAK_READ)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  ansPtr = BDI_ExtractLong(count, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY (%lu)\n", *count);
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_BreakRead */


/****************************************************************************
 ****************************************************************************

    Delete one or all software breakpoint(s).

    INPUT:  breakId       the ID returned by the BDI_BreakAdd() function
                          or -1 to delete all watchpoints
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_BreakDelete(int breakId)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_BreakDelete(%i)", breakId);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_BREAK_DELETE, cmdBuffer);
  if (breakId < 0) cmdPtr = BDI_AppendWord(0xFFFF, cmdPtr);
  else             cmdPtr = BDI_AppendWord((WORD)breakId, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_BREAK_DELETE)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_BreakDelete */


/****************************************************************************
 ****************************************************************************

  Add a hardware breakpoint (watch).

    INPUT:  mode          the watch mode
            from          the first address to watch
            to            the last address to watch
            count         number of matches before break
            data          the data pattern to compare
    OUTPUT: RETURN        watch identification or a negativ error code

 ****************************************************************************/

int BDI_WatchAdd(WORD mode, DWORD from, DWORD to, DWORD count, DWORD data)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      watchId;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_WatchAdd(%u, 0x%08lx, 0x%08lx, %lu, 0x%08lx)", mode, from, to, count, data);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_WATCH_ADD, cmdBuffer);
  cmdPtr = BDI_AppendWord(mode,  cmdPtr);
  cmdPtr = BDI_AppendLong(from,  cmdPtr);
  cmdPtr = BDI_AppendLong(to,    cmdPtr);
  cmdPtr = BDI_AppendLong(count, cmdPtr);
  cmdPtr = BDI_AppendLong(data,  cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 5) || (answer != BDI_CMD_WATCH_ADD)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  if (errorCode == 0) {
    ansPtr = BDI_ExtractWord(&watchId, ansPtr);
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY (id=%i)\n", watchId);
    } /* if */
#endif
    return (int)watchId;
  } /* if */
  else {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
#endif
    return (int)(short)errorCode;
  } /* else */
} /* BDI_WatchAdd */


/****************************************************************************
 ****************************************************************************

    Delete one or all watchpoint(s).

    INPUT:  watchId       the ID returned by the BDI_WatchAdd() function
                          or -1 to delete all watchpoints
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_WatchDelete(int watchId)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_WatchDelete(%i)", watchId);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_WATCH_DELETE, cmdBuffer);
  if (watchId < 0) cmdPtr = BDI_AppendWord(0xFFFF, cmdPtr);
  else             cmdPtr = BDI_AppendWord((WORD)watchId, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_WATCH_DELETE)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_WatchDelete */


/****************************************************************************
 ****************************************************************************

    Get info about the BDI firmware and the current loaded setup

    INPUT:  -
    OUTPUT: bdiInfo       the BDI info structure
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_InfoGet(BDI_InfoT *bdiInfo)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  char      bdiType;
  WORD      version;
  WORD      cpuType;
  DWORD     time;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_InfoGet()");
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_BDI_INFO_GET, cmdBuffer);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount > (10 + BDI_MAX_ID_SIZE)) || (answer != BDI_CMD_BDI_INFO_GET)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractChar(&bdiType, ansPtr);
  ansPtr = BDI_ExtractWord(&version, ansPtr);
  ansPtr = BDI_ExtractWord(&cpuType, ansPtr);
  ansPtr = BDI_ExtractLong(&time,    ansPtr);
  ansBuffer[rxCount] = 0; // for security
  strcpy(bdiInfo->setupId, (char*)ansPtr);

  /* fill info structure */
  if (bdiType == 'F') {
    bdiInfo->fwType = (version >>  8);
    bdiInfo->fwVersion = (version & 0x00FF);
  } /* if */
  else if (version >= 0xC000) {
    version -= 0xC000;
    bdiInfo->fwType = (version >>  8);
    bdiInfo->fwVersion = (version & 0x00FF);
  } /* if */
  else {
    bdiInfo->fwType = (version >> 12);
    bdiInfo->fwVersion = (version & 0x0FFF);
  } /* else */
  bdiInfo->cpuType = cpuType;
  bdiInfo->setupTimeStamp = time;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_InfoGet */


/****************************************************************************
 ****************************************************************************

    Get info about the BDI type and the loaded firmware

    INPUT:  -
    OUTPUT: bdiType       BDI type
            fwType        firmware type
            fwVersion     firmware version
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_TypeGet(WORD* bdiType, WORD* fwType, WORD* fwVersion)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  char      bdiChar;
  WORD      version;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_TypeGet()");
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_BDI_INFO_GET, cmdBuffer);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount > (10 + BDI_MAX_ID_SIZE)) || (answer != BDI_CMD_BDI_INFO_GET)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractChar(&bdiChar, ansPtr);
  ansPtr = BDI_ExtractWord(&version, ansPtr);

  /* return result */
  if      (bdiChar == 'H') *bdiType = BDI_TYPE_BDIHS;
  else if (bdiChar == 'C') *bdiType = BDI_TYPE_BDI1000;
  else if (bdiChar == 'E') *bdiType = BDI_TYPE_BDI2000;
  else if (bdiChar == 'F') *bdiType = BDI_TYPE_BDI3000;
  else                     *bdiType = BDI_TYPE_UNKOWN;

  if (*bdiType == BDI_TYPE_BDI3000) {
    *fwType = (version >>  8);
    *fwVersion = (version & 0x00FF);
  } /* if */
  else if (version >= 0xC000) {
    version -= 0xC000;
    *fwType = (version >>  8);
    *fwVersion = (version & 0x00FF);
  } /* if */
  else {
    *fwType = (version >> 12);
    *fwVersion = (version & 0x0FFF);
  } /* else */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_TypeGet */


/****************************************************************************
 ****************************************************************************

    Get the current target state

    INPUT:  -
    OUTPUT: state         the current target state
            error         the last detected error before entering error state
            timeStamp     the reset time stamp
            count         number of additional information fields
            info          the additional infos
            RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_StateGet(WORD  *state,
                 int   *error,
                 DWORD *timeStamp,
                 WORD  *count,
                 DWORD *info)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  BYTE      targetState;
  BYTE      infoCount;
  WORD      errorCode;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_StateGet()");
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_TARGET_STATE_GET, cmdBuffer);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            3000);  /* because of long breakpoint inser time */
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if (answer != BDI_CMD_TARGET_STATE_GET) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractByte(&targetState, ansPtr);
  *state = targetState;
  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  *error = (int)(short)errorCode;
  ansPtr = BDI_ExtractLong(timeStamp, ansPtr);
  ansPtr = BDI_ExtractByte(&infoCount, ansPtr);
  if (infoCount > *count) infoCount = (BYTE)(*count);
  *count = infoCount;
  while (infoCount--) {
    ansPtr = BDI_ExtractLong(info++, ansPtr);
  } /* while */

#ifdef _DEBUG
  if (logFile != NULL) {
    if      (targetState == 0) fprintf(logFile, " => OKAY (reseted)\n");
    else if (targetState == 1) fprintf(logFile, " => OKAY (freezed)\n");
    else if (targetState == 2) fprintf(logFile, " => OKAY (running)\n");
    else if (targetState == 3) fprintf(logFile, " => OKAY (error = %s)\n", BDI_ErrorTextGet(*error));
    else                       fprintf(logFile, " => OKAY (unknown)\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_StateGet */


/****************************************************************************
 ****************************************************************************

  Read Terminal Output characters from the target.

    INPUT:  -
    OUTPUT: data          the characters read from the target
            RETURN        number or read characters or a negativ error code

 ****************************************************************************/

int BDI_TerminalRead(char *data)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_TerminalRead()");
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_GET_TERMINAL, cmdBuffer);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if (answer != BDI_ANS_WRITE_TERMINAL) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  /* copy received data to user buffer */
  ansBuffer[rxCount] = 0; // for security
  strcpy(data, (char*)ansPtr);

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY (%s)\n", data);
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_TerminalRead */


/****************************************************************************
 ****************************************************************************

  Write a character to the target terminal channel. There is no buffering
  within the BDI. If the target buffer is not free, the character is lost.

    INPUT:  ch            the character to write
    OUTPUT: RETURN        number or read characters or a negativ error code

 ****************************************************************************/

int BDI_TerminalWrite(char ch)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_TerminalWrite(%c)", ch);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_READ_TERMINAL, cmdBuffer);
  cmdPtr = BDI_AppendChar(ch, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 1) || (answer != BDI_ANS_ACK)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_TerminalWrite */


/****************************************************************************
 ****************************************************************************

  Sends a command string to the BDI and waits for the answer.

    INPUT  : pszCommand   points to the command string
             nBufferSize  the size of the answer buffer
             pszAnswer    points to the answer string
    OUTPUT : RETURN       0 if okay or a negative number if error

 ****************************************************************************/

int BDI_DirectCommand(const char *pszCommand,
                            WORD  nBufferSize,
                            char *pszAnswer)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  char     *putPtr;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_DirectCommand(%s)", pszCommand);
  } /* if */
#endif

  /* check command size */
  if (strlen(pszCommand) >= BDI_MAX_BLOCK_SIZE) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_DIRECT_COMMAND, cmdBuffer);
  do {
    *cmdPtr++ = (BYTE)(*pszCommand);
  } while (*pszCommand++);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            5000);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if (answer != BDI_CMD_DIRECT_COMMAND) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  /* copy answer */
  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  ansBuffer[rxCount] = 0; // for security
  putPtr = pszAnswer;
  while ((*ansPtr != 0) && (nBufferSize > 1)) {
    *putPtr++ = (char)(*ansPtr++);
    nBufferSize--;
  } /* while */
  *putPtr = 0;

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY (%s)\n", pszAnswer);
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_DirectCommand */


/****************************************************************************
 ****************************************************************************

    Read from a debug channel

    INPUT:  channel       the channel ID
            count         number of bytes to read
    OUTPUT: data          the read data will be stored in this buffer
            RETURN        the number of bytes read or a negativ number if error

 ****************************************************************************/

int BDI_ChannelRead(WORD channel, WORD count, BYTE *data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  WORD      currentCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      readCount;
  WORD      totalCount;
  WORD      i;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_ChannelRead(%u, %u)", channel, count);
  } /* if */
#endif

  totalCount = 0;
  while (count > 0) {

    /* get next partial block */
    if (count > BDI_MAX_BLOCK_SIZE) currentCount = BDI_MAX_BLOCK_SIZE;
    else                            currentCount = count;

    /* prepare command */
    cmdPtr = BDI_AppendByte(BDI_CMD_CHANNEL_READ, cmdBuffer);
    cmdPtr = BDI_AppendWord(channel,      cmdPtr);
    cmdPtr = BDI_AppendWord(currentCount, cmdPtr);
    count -= currentCount;

    /* BDI transaction */
    rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                              cmdBuffer,
                              sizeof ansBuffer,
                              ansBuffer,
                              1000);
    if (rxCount < 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
      } /* if */
#endif
      return rxCount;
    } /* if */

    /* analyse response */
    ansPtr = ansBuffer;
    answer = *ansPtr++;
    if (answer != BDI_CMD_CHANNEL_READ) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
      } /* if */
#endif
      return BDI_ERR_INVALID_RESPONSE;
    } /* if */

    ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
    if (errorCode != 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
      } /* if */
#endif
      return (int)(short)errorCode;
    } /* if */

    ansPtr = BDI_ExtractWord(&readCount, ansPtr);
    for (i = 0; i < readCount; i++) *data++ = *ansPtr++;
    totalCount += readCount;
    if (readCount < currentCount) break;
  } /* while */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY (%u)\n", totalCount);
  } /* if */
#endif
  return (int)totalCount;
} /* BDI_ChannelRead */


/****************************************************************************
 ****************************************************************************

    Write to a debug channel

    INPUT:  channel       the channel ID
            count         number of bytes to write
            data          data to write to the target memory
    OUTPUT: RETURN        number of bytes written or a negativ number if error

 ****************************************************************************/

int BDI_ChannelWrite(WORD channel, WORD count, const BYTE *data)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  WORD      currentCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      written;
  WORD      totalCount;
  WORD      i;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_ChannelWrite(%u, %u)", channel, count);
  } /* if */
#endif

  totalCount = 0;
  while (count > 0) {

    /* get next partial block */
    if (count > BDI_MAX_BLOCK_SIZE) currentCount = BDI_MAX_BLOCK_SIZE;
    else                            currentCount = count;

    /* prepare command */
    cmdPtr = BDI_AppendByte(BDI_CMD_CHANNEL_WRITE, cmdBuffer);
    cmdPtr = BDI_AppendWord(channel,      cmdPtr);
    cmdPtr = BDI_AppendWord(currentCount, cmdPtr);
    for (i=0; i<currentCount; i++) *cmdPtr++ = *data++;
    count -= currentCount;

    /* BDI transaction */
    rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                              cmdBuffer,
                              sizeof ansBuffer,
                              ansBuffer,
                              1000);
    if (rxCount < 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
      } /* if */
#endif
      return rxCount;
    } /* if */

    /* analyse response */
    ansPtr = ansBuffer;
    answer = *ansPtr++;
    if (answer != BDI_CMD_CHANNEL_WRITE) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
      } /* if */
#endif
      return BDI_ERR_INVALID_RESPONSE;
    } /* if */

    ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
    if (errorCode != 0) {
#ifdef _DEBUG
      if (logFile != NULL) {
        fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
      } /* if */
#endif
      return (int)(short)errorCode;
    } /* iff */

    ansPtr = BDI_ExtractWord(&written, ansPtr);
    totalCount += written;
    if (written < currentCount) break;
  } /* while */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY (%u)\n", totalCount);
  } /* if */
#endif
  return (int)totalCount;
} /* BDI_ChannelWrite */


/****************************************************************************
 ****************************************************************************

    Select the core to work with

    INPUT:  core          the core number
            option        an option
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_CoreSelect(WORD core, DWORD option)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  BYTE      answer;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_CoreSelect(%u, 0x%08lx)", core, option);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_CORE_SELECT, cmdBuffer);
  cmdPtr = BDI_AppendWord(core, cmdPtr);
  cmdPtr = BDI_AppendLong(option, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 1) || (answer != BDI_CMD_CORE_SELECT)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_CoreSelect */


/****************************************************************************
 ****************************************************************************

  Returns an error text to a BDI error number.

    INPUT  : error      the BDI error number
    OUTPUT : RETURN     pointer to error text

 ****************************************************************************/

static const struct {int nbr; char* text;} ErrorTable[] = {

{ BDI_OKAY                  , "Okay" },

{ BDI_ASYN_SETUP            , "LNK: Communication setup failed (Port busy)" },
{ BDI_ASYN_RX_TIMEOUT       , "LNK: Asynchronous receive timeout" },
{ BDI_ASYN_RX_ERROR         , "LNK: Asynchronous receive error" },
{ BDI_ASYN_RX_FORMAT        , "LNK: Asynchronous receive format error" },
{ BDI_ASYN_RX_BCC           , "LNK: Asynchronous receive BCC error" },
{ BDI_ASYN_RX_OVERFLOW      , "LNK: Asynchronous receive overflow" },
{ BDI_ASYN_TX_ERROR         , "LNK: Asynchronous transmit error" },

{ BDI_SOCKET_ERROR          , "LNK: Network socket error" },
{ BDI_SOCKET_RX_TIMEOUT     , "LNK: Network receive error" },

{ BDI_ERR_NO_RESPONSE       , "BDI: No response from BDI" },
{ BDI_ERR_INVALID_PARAMETER , "BDI: Invalid function parameter" },
{ BDI_ERR_NOT_CONNECTED     , "BDI: BDI not connected" },
{ BDI_ERR_ANSWER_TOO_BIG    , "BDI: BDI answer too big" },
{ BDI_ERR_INVALID_RESPONSE  , "BDI: Invalid response received from BDI" },
{ BDI_ERR_MEM_ACCESS        , "BDI: Download failed" },
{ BDI_ERR_FILE_ACCESS       , "BDI: File access failed" },

{ BDI_ERR_FLASH_ERASE       , "BDI: Erasing flash sector failed" },
{ BDI_ERR_FLASH_PROGRAM     , "BDI: Programming flash failed" },
{ BDI_ERR_VERIFY            , "BDI: Verifying memory failed" },
{ BDI_ERR_INVALID_MODE      , "BDI: Invalid write mode selected" },
{ BDI_ERR_WORKING_RAM       , "BDI: Cannot write to working RAM" },
{ BDI_ERR_USER_ABORT        , "BDI: Download aborted by user" },

{ BDI_ERR_UNKNOWN_BDI       , "SETUP: Unknown BDI connected" },
{ BDI_ERR_FIRMWARE_FILE     , "SETUP: Access to firmware file failed" },
{ BDI_ERR_FLASH_VERIFY      , "SETUP: Verify of firmware failed" },
{ BDI_ERR_LOGIC_DEVICE      , "SETUP: Cannot read logic device ID" },
{ BDI_ERR_LOGIC_VERIFY      , "SETUP: Verify of logic failed" },
{ BDI_ERR_LOGIC_FILE        , "SETUP: Access to logic file failed" },

{ -2001, "BDM: Communication failed" },
{ -2002, "BDM: Target Reset always asserted" },
{ -2003, "BDM: Invalid target connected" },
{ -2004, "BDM: Timeout while waiting for freeze" },
{ -2005, "BDM: Target reset detected" },
{ -2006, "BDM: Wrong target state for requested command " },
{ -2007, "BDM: Unexpected response from target" },
{ -2008, "BDM: Memory test failed" },
{ -2009, "BDM: Target has no power" },
{ -2010, "BDM: Bus error - Terminated bus cycle" },

{ -2011, "HC12: No debug clock detected (ECLK)" },

{ -2021, "SAP: Read access failed" },
{ -2022, "SAP: Write access failed" },
{ -2023, "Nexus: Read access failed" },
{ -2024, "Nexus: Write access failed" },
{ -2025, "SAP: HReset detected" },
{ -2026, "SAP: Overflow error" },
{ -2027, "SAP: Underflow error" },
{ -2028, "SAP: XFER error" },
{ -2029, "SAP: Security error" },

{ -2031, "PPC: Target interrupt indication received (may be bus error)" },
{ -2032, "PPC: Sequencing error received" },
{ -2033, "PPC: Unexpected data received" },
{ -2034, "PPC: Data expected but not received" },
{ -2035, "PPC: This target is not supported" },
{ -2036, "PPC: Writing to workspace failed" },
{ -2037, "PPC: Instruction stuff overrun" },
{ -2038, "PPC: MMU translation error" },
{ -2039, "PPC: OnCE single-step error (may be bus error)" },

{ -2041, "ARM: Data abort (bus error)" },
{ -2042, "ARM: This request is not supported" },
{ -2043, "ARM: Unknown target CPU type" },
{ -2044, "XSCALE: Communication with debug handler failed" },
{ -2045, "XSCALE: High speed download overrun" },
{ -2046, "ARM: Writing to workspace failed" },

{ -2051, "EJTAG: Bus error while in debug mode" },
{ -2052, "EJTAG: TLB exception while in debug mode" },
{ -2053, "EJTAG: Timeout while waiting for data access" },
{ -2054, "EJTAG: Misaligned data access (address error)" },
{ -2055, "EJTAG: Writing to workspace failed" },
{ -2056, "EJTAG: Timeout while waiting for opcode fetch" },
{ -2057, "EJTAG: Exception in debug mode detected" },

{ -2061, "DAP: fatal communication error" },
{ -2062, "DAP: timeout during debug access" },
{ -2063, "DAP: error during debug access" },
{ -2064, "DAP: overrun during debug access" },
{ -2065, "DAP: no CoreSight debug component found" },

{ -2071, "JTAG: Controller test failed" },
{ -2072, "JTAG: Loopback test failed" },
{ -2073, "JTAG: INT pin test failed" },
{ -2074, "JTAG: Communication test failed" },
{ -2075, "JTAG: Communication timeout detected" },
{ -2076, "JTAG: No communication clock detected" },
{ -2077, "JTAG: JTAG lockout recovery faild" },

{ -2081, "JIO: Error writing instruction" },
{ -2082, "JIO: No valid start bit received" },
{ -2083, "JIO: Writing scan data faild" },
{ -2084, "JIO: Reading scan data failed" },
{ -2085, "JIO: Setting address failed" },
{ -2086, "JIO: Dirty bit received" },

{ TARGET_ERR_STATE,         "BDI: Target must be in debug mode for this action" },
{ TARGET_ERR_PARAMETER,     "BDI: Invalid parameter" },
{ TARGET_ERR_STEP_TIMEOUT,  "BDI: Step timeout detected" },
{ TARGET_ERR_RESOURCE,      "BDI: Not enough resources for this command" },
{ TARGET_ERR_VERIFY,        "BDI: Verifying target memory failed" },
{ TARGET_ERR_WRITE,         "BDI: Writing to target memory failed" },
{ -2107, "BDI: Invalid target memory address" },
{ -2109, "BDI: Loading code to BDI workspace faild" },

{ -2201, "Direct Cmd: Unknown Object" },
{ -2202, "Direct Cmd: Unknown Action" },
{ -2203, "Direct Cmd: Unknown Parameter" },
{ -2204, "Direct Cmd: Invalid BDI Setup" },
{ -2205, "Direct Cmd: Invalid Parameter Value" },
{ -2206, "Direct Cmd: Not supported for this target" },
{ -2207, "Direct Cmd: Execution failed" }

};

static char szUnknown[80];

const char* BDI_ErrorTextGet(int error)
{
  int   i;
  char* textPtr;

  /* find text */
  textPtr = NULL;
  for (i = 0; i < (int)(sizeof ErrorTable / sizeof ErrorTable[0]); i++) {
    if (ErrorTable[i].nbr == error) {
      textPtr = ErrorTable[i].text;
      break;
    } /* if */
  } /* for */

  if (textPtr != NULL) {
    return textPtr;
  } /* if */
  else {
    sprintf(szUnknown, "BDI: Unknown error number (%i)", error);
    return szUnknown;
  } /* else */


} /* BDI_ErrorTextGet */


/****************************************************************************
 ****************************************************************************
                Additional BDI Access functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    Reset and startup target with an init list

    INPUT:  resetTime     reset time in milliseconds
            cpuClock      CPU clock in hertz after the init list is processed
            initCount     number of entries in the init list
            pInitList     points to the initialisation list
    OUTPUT: RETURN        0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_TargetStartup(      WORD           resetTime,
                            DWORD          cpuClock,
                            WORD           initCount,
                      const BDI_InitTypeT *pInitList)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  DWORD     executionTime;

#ifdef _DEBUG
  WORD           i;
  const BDI_InitTypeT *pInit;

  if (logFile != NULL) {
    fprintf(logFile, "BDI_TargetStartup(%u, %lu, %u)\n", resetTime, cpuClock, initCount);
    pInit = pInitList;
    for (i = 0; i < initCount; i++) {
      fprintf(logFile, "  init: %2li, 0x%08lx, 0x%08lx\n",
              pInit->command, pInit->address, pInit->value);
      pInit++;
    } /* if */
  } /* if */
#endif

  /* check parameters */
  if (initCount > 128) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_TARGET_STARTUP, cmdBuffer);
  cmdPtr = BDI_AppendWord(resetTime, cmdPtr);
  cmdPtr = BDI_AppendLong(cpuClock,  cmdPtr);
  cmdPtr = BDI_AppendWord(initCount, cmdPtr);
  while (initCount > 0) {
    cmdPtr = BDI_AppendWord((WORD)pInitList->command, cmdPtr);
    cmdPtr = BDI_AppendLong(pInitList->address, cmdPtr);
    cmdPtr = BDI_AppendLong(pInitList->value,   cmdPtr);
    initCount--;
    pInitList++;
  } /* while */

  /* BDI transaction */
  if (resetTime > 0) executionTime = resetTime + 3000;
  else               executionTime = 5000;
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            executionTime);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_CMD_TARGET_STARTUP)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  memWriteMode = BDI_LOAD_MODE_RAM; /* write mode has changed to RAM */
  pagedAddress = FALSE;

#ifdef _DEBUG
  if (errorCode == 0) {
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
  } /* if */
  else {
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
  } /* else */
#endif

  return (int)(short)errorCode;
} /* BDI_TargetStartup */


/****************************************************************************
 ****************************************************************************

    Sets the speed for the BDM communication.

    The new speed is used until the next target reset

    INPUT  : cpuClock     the clock rate of the target processor
    OUTPUT : return       0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_SetBdmSpeed(DWORD cpuClock)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_SetBdmSpeed(%lu)", cpuClock);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_SET_BDM_SPEED, cmdBuffer);
  cmdPtr = BDI_AppendLong(cpuClock, cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 1) || (answer != BDI_CMD_SET_BDM_SPEED)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_SetBdmSpeed */


/****************************************************************************
 ****************************************************************************

    Writes an option string to the BDI

    INPUT  : id           the option ID
             data         the option string
    OUTPUT : return       0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_SetOption(BYTE id, const char* data)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  size_t    count;

#ifdef _DEBUG
  if (logFile != NULL) {
    if (data != NULL) fprintf(logFile, "BDI_SetOption(%u, %s)", id, data);
    else              fprintf(logFile, "BDI_SetOption(%u, NULL)", id);
  } /* if */
#endif

  /* check string length */
  count = 0;
  if (data != NULL) count = strlen(data);
  if (count > BDI_MAX_BLOCK_SIZE) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_SET_OPTION, cmdBuffer);
  cmdPtr = BDI_AppendByte(id, cmdPtr);
  cmdPtr = BDI_AppendWord((WORD)count, cmdPtr);
  while (count--) *cmdPtr++ = *data++;

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 1) || (answer != BDI_CMD_SET_OPTION)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_SetOption */


/****************************************************************************
 ****************************************************************************

    Reads a target byte,word,long

    INPUT:  addr        address of the target
    OUTPUT: value       the read value
            return      0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_GetByte(DWORD addr, BYTE *value)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_GetByte(0x%08lx)", addr);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_FETCH_BYTE, cmdBuffer);
  cmdPtr = BDI_AppendLong(addr,   cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 2) || (answer != BDI_ANS_SEND_BYTE)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractByte(value, ansPtr);
#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY (0x%02x)\n", *value);
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_GetByte */

int BDI_GetWord(DWORD addr, WORD *value)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_GetWord(0x%08lx)", addr);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_FETCH_WORD, cmdBuffer);
  cmdPtr = BDI_AppendLong(addr,   cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_ANS_SEND_WORD)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(value, ansPtr);
#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY (0x%04x)\n", *value);
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_GetWord */

int BDI_GetLong(DWORD addr, DWORD *value)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_GetLong(0x%08lx)", addr);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_FETCH_LONG, cmdBuffer);
  cmdPtr = BDI_AppendLong(addr,   cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 5) || (answer != BDI_ANS_SEND_LONG)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractLong(value, ansPtr);
#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY (0x%08lx)\n", *value);
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_GetLong */


/****************************************************************************
 ****************************************************************************

    Writes to a target byte,word,long

    INPUT:  addr            address of the target
            value           value to write to the target
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_SetByte(DWORD addr, BYTE  value)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  BYTE      rxValue;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_SetByte(0x%08lx, 0x%02x)", addr, value);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_SET_BYTE, cmdBuffer);
  cmdPtr = BDI_AppendLong(addr,   cmdPtr);
  cmdPtr = BDI_AppendByte(value,  cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 2) || (answer != BDI_ANS_SEND_BYTE)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractByte(&rxValue, ansPtr);
  if (rxValue != value) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(TARGET_ERR_WRITE));
    } /* if */
#endif
    return TARGET_ERR_WRITE;
  } /* if */
  else {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
#endif
    return BDI_OKAY;
  } /* else */
} /* BDI_SetByte */


int BDI_SetWord(DWORD addr, WORD  value)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      rxValue;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_SetWord(0x%08lx, 0x%04x)", addr, value);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_SET_WORD, cmdBuffer);
  cmdPtr = BDI_AppendLong(addr,   cmdPtr);
  cmdPtr = BDI_AppendWord(value,  cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 3) || (answer != BDI_ANS_SEND_WORD)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractWord(&rxValue, ansPtr);
  if (rxValue != value) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(TARGET_ERR_WRITE));
    } /* if */
#endif
    return TARGET_ERR_WRITE;
  } /* if */
  else {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
#endif
    return BDI_OKAY;
  } /* else */
} /* BDI_SetWord */


int BDI_SetLong(DWORD addr, DWORD value)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  DWORD     rxValue;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_SetWord(0x%08lx, 0x%08lx)", addr, value);
  } /* if */
#endif

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_SET_LONG, cmdBuffer);
  cmdPtr = BDI_AppendLong(addr,   cmdPtr);
  cmdPtr = BDI_AppendLong(value,  cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            BDI_DEFAULT_EXEC_TIME);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 5) || (answer != BDI_ANS_SEND_LONG)) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  ansPtr = BDI_ExtractLong(&rxValue, ansPtr);
  if (rxValue != value) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(TARGET_ERR_WRITE));
    } /* if */
#endif
    return TARGET_ERR_WRITE;
  } /* if */
  else {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => OKAY\n");
    } /* if */
#endif
    return BDI_OKAY;
  } /* else */
} /* BDI_SetLong */


/****************************************************************************
 ****************************************************************************

    Reads a block of bytes from the target memory

    INPUT:  addr            address of the target
            count           number of bytes to read (up to 64k)
    OUTPUT: block           the read data block
            return          error code

 ****************************************************************************/

int BDI_DumpBlock(DWORD addr, WORD count, BYTE *block)
{
  int   result;
  WORD  space;

  /* enable paging */
  result = BDI_OKAY;
  if ((targetIsHC12) || (targetIsS12X)) {
    memWriteMode = BDI_LOAD_MODE_RAM;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, cpuTypeHC12);
  } /* if */

  /* dump block */
  space = (targetIsS12X) ? BDI_AT_GLOBAL : BDI_AT_DEFAULT;
  if (result == BDI_OKAY) {
    result = BDI_MemoryRead(addr, space, count, block);
  } /* if */

  return result;
} /* BDI_DumpBlock */


/****************************************************************************
 ****************************************************************************

    Reads a block of bytes from the target memory and stores it in a
    S-Record file.

    INPUT:  fileName        the full path and filename of the S-record file
            addr            address of the target
            count           number of bytes to read
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_DumpFile(const char *fileName, DWORD addr, DWORD count)
{
  int       result;
  FILE*     file;
  char      szLine[100];
  BYTE      memBlock[32];
  WORD      memCount;
  size_t    lineLength;
  char      chDataType;
  char      chTermType;
  WORD      space;
  DWORD     totalCount;

  /* open S-Record file */
  file = fopen(fileName,"w");
  if (file == NULL) return BDI_ERR_FILE_ACCESS;

  /* determine S record type */
  if ((addr + count - 1) <= 0xFFFF)
  {
    chDataType = '1';
    chTermType = '9';
  }
  else if ((addr + count - 1) <= 0xFFFFFF)
  {
    chDataType = '2';
    chTermType = '8';
  }
  else
  {
    chDataType = '3';
    chTermType = '7';
  }

  /* enable paging */
  result = BDI_OKAY;
  if ((targetIsHC12) || (targetIsS12X)) {
    memWriteMode = BDI_LOAD_MODE_RAM;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, cpuTypeHC12);
  } /* if */

  /* read target memory */
  totalCount = 0;
  space = (targetIsS12X) ? BDI_AT_GLOBAL : BDI_AT_DEFAULT;
  while ((count > 0) && (result == BDI_OKAY)) {
    if (count > sizeof memBlock) memCount = sizeof memBlock;
    else                         memCount = (WORD)count;
    result = BDI_MemoryRead(addr, space, memCount, memBlock);
    if (result == BDI_OKAY) {
      lineLength = BuildSRecord(chDataType, addr, memCount, memBlock, szLine);
      addr  += memCount;
      count -= memCount;
      if (fwrite(szLine, sizeof(char), lineLength, file) != lineLength) {
        result = BDI_ERR_FILE_ACCESS;
      } /* if */

      /* progress callback */
      totalCount += memCount;
      if (progressCallback != NULL) {
        if (progressCallback(totalCount) == BDI_ABORT_REQUEST) result = BDI_ERR_USER_ABORT;
      } /* if */

    } /* if */
  } /* while */

  /* close file */
  fclose(file);

  return result;
} /* BDI_DumpFile */


/****************************************************************************
 ****************************************************************************

    Reads a block of bytes from the target memory and stores it in a
    Binary file.

    INPUT:  fileName        the full path and filename of the Binary file
            addr            address of the target
            count           number of bytes to read
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_DumpBinary(const char *fileName, DWORD addr, DWORD count)
{
  int       result;
  FILE*     file;
  BYTE      memBlock[BDI_MAX_BLOCK_SIZE];
  WORD      memCount;
  WORD      space;
  DWORD     totalCount;

  /* open binary file */
  file = fopen(fileName,"wb");
  if (file == NULL) return BDI_ERR_FILE_ACCESS;

  /* enable paging */
  result = BDI_OKAY;
  if ((targetIsHC12) || (targetIsS12X)) {
    memWriteMode = BDI_LOAD_MODE_RAM;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, cpuTypeHC12);
  } /* if */

  /* read target memory */
  totalCount = 0;
  space = (targetIsS12X) ? BDI_AT_GLOBAL : BDI_AT_DEFAULT;
  while ((count > 0) && (result == BDI_OKAY)) {
    if (count > sizeof memBlock) memCount = sizeof memBlock;
    else                         memCount = (WORD)count;
    result = BDI_MemoryRead(addr, space, memCount, memBlock);
    if (result == BDI_OKAY) {
      addr  += memCount;
      count -= memCount;
      if (fwrite(memBlock, 1, memCount, file) != memCount) {
        result = BDI_ERR_FILE_ACCESS;
      } /* if */

      /* progress callback */
      totalCount += memCount;
      if (progressCallback != NULL) {
        if (progressCallback(totalCount) == BDI_ABORT_REQUEST) result = BDI_ERR_USER_ABORT;
      } /* if */

    } /* if */
  } /* while */

  /* close file */
  fclose(file);

  return result;
} /* BDI_DumpBinary */


/****************************************************************************
 ****************************************************************************

    Writes a block to the target memory
    The Write mode of the BDI is already set.

    INPUT:  addr            address of the target
            count           number of bytes in the block (up to 64k)
            block           the data block
    OUTPUT: errorAddr       address of the failing byte
            return          error code

 ****************************************************************************/

static int BDI_WriteBlock(DWORD   addr,
                          WORD    count,
                          BYTE   *block,
                          DWORD  *errorAddr)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  WORD      maxCount;
  WORD      currentCount;
  BYTE      answer;
  BYTE      loadResult;
  WORD      i;

  /* use small blocks for internal EEPROM */
  if (    (memWriteMode == BDI_HC12_EEPROM)
       || (memWriteMode == BDI_HC12A_EEPROM)
       || (memWriteMode == BDI_STAR12_EEPROM)
     ) {
    maxCount = 32;
  } /* if */
  else {
    maxCount = BDI_MAX_BLOCK_SIZE;
  } /* else */

  while (count > 0) {

    /* get next partial block, use multiple of 4 for max block */
    if (count > maxCount) currentCount = maxCount;
    else                  currentCount = count;

    /* prepare command */
    cmdPtr = BDI_AppendByte(BDI_CMD_LOAD_SINGLE_BLOCK, cmdBuffer);
    cmdPtr = BDI_AppendLong(addr,         cmdPtr);
    cmdPtr = BDI_AppendWord(currentCount, cmdPtr);
    for (i=0; i<currentCount; i++) *cmdPtr++ = *block++;

    /* BDI transaction */
    rxCount = BDI_Transaction(cmdPtr-cmdBuffer, cmdBuffer, sizeof ansBuffer, ansBuffer, 200 + currentCount);
    if (rxCount < 0) return rxCount;

    /* analyse response */
    ansPtr = BDI_ExtractByte(&answer, ansBuffer);
    if ((rxCount != 6) || (answer != BDI_CMD_LOAD_SINGLE_BLOCK))
      return BDI_ERR_INVALID_RESPONSE;

    ansPtr = BDI_ExtractByte(&loadResult, ansPtr);
    ansPtr = BDI_ExtractLong(errorAddr,   ansPtr);
    if (loadResult != 0) return BDI_ERR_MEM_ACCESS;

    /* adjust count and address for the next partial block */
    count -= (WORD)currentCount;
    addr += currentCount;

  } /* while */

  return BDI_OKAY;
} /* BDI_WriteBlock */


/****************************************************************************
 ****************************************************************************

    Writes a S-Record file into the target memory
    The Write mode of the BDI is already set.

    INPUT:  fileName        the full path and filename of the S-record file
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

static BYTE      lineData[SREC_MAX_DATA];
static int       lineCount;
static DWORD     lineAddr;

static int BDI_ReadNextFileBlock(FILE* file, BYTE* data, DWORD* addr)
{
  int       result = BDI_OKAY;
  char      szLine[SREC_MAX_LINE];
  int       blockCount;
  DWORD     blockAddr;

  /* append last line */
  blockCount = lineCount;
  if (lineCount > 0)
  {
    blockCount = lineCount;
    blockAddr  = lineAddr;
    memcpy(data, lineData, lineCount);
    lineCount  = 0;
  }

  /* process next line, append to buffer if continous */
  while ((result == BDI_OKAY) && (fgets(szLine, sizeof szLine, file) != NULL))
  {
    lineCount = DecodeSRecord(szLine, &lineAddr, lineData);
    if (lineCount > 0)
    {
	  if (blockCount == 0) blockAddr = lineAddr;
	  if (   ((blockAddr + blockCount) == lineAddr)
	      && ((blockCount + lineCount) <= BDI_MAX_BLOCK_SIZE))
	  {
	    memcpy(data + blockCount, lineData, lineCount);
		blockCount += lineCount;
		lineCount   = 0;
	  } /* if */
	  else
	  {
	    break;
	  }
    } /* if */
    else if (lineCount < 0)
    {
      result = BDI_ERR_FILE_ACCESS;
    }
  }
  *addr = blockAddr;
  if (result == BDI_OKAY) return blockCount;
  else                    return result;
} /* BDI_ReadNextFileBlock */

static int BDI_WriteFile(const char *fileName, DWORD offset, DWORD *errorAddr)
{
  int       result = BDI_OKAY;
  FILE*     file;
  BYTE      blockData[BDI_MAX_BLOCK_SIZE];
  int       blockCount;
  DWORD     totalCount;
  DWORD     blockAddr;
  DWORD     pageAddr;
  DWORD     linAddr;
  WORD      splitCount;
  BYTE*     pData;

  /* open S-Record file */
  file = fopen(fileName,"r");
  if (file == NULL) return BDI_ERR_FILE_ACCESS;

  /* init download */
  result = BDI_InitDownload(BDI_NO_VERIFY, TRUE);

  /* download data */
  lineCount  = 0;
  totalCount = 0;
  do {
    blockCount = BDI_ReadNextFileBlock(file, blockData, &blockAddr);
    if (blockCount > 0) {
      totalCount += (DWORD)blockCount;
      blockAddr  += offset;

      /* S12X: build global Address */
      if (targetIsS12X && pagedAddress) {
        if ((blockAddr & 0xC000) == 0x8000) {
          linAddr  = (blockAddr & 0xFFFF0000) >> 2;
          linAddr += (blockAddr & 0x3FFF);
          linAddr += 0x400000;
        } /* if */
        else if ((blockAddr & 0xC000) == 0xC000) {
          linAddr = 0x7FC000 + (blockAddr & 0x3FFF);
        } /* else if */
        else if ((blockAddr & 0xC000) == 0x4000) {
          linAddr = 0x7F4000 + (blockAddr & 0x3FFF);
        } /* else if */
        else if ((blockAddr & 0xFC00) == 0x0800) {
          linAddr  = (blockAddr & 0xFFFF0000) >> 6;
          linAddr += (blockAddr & 0x03FF);
          linAddr += 0x100000;
        } /* else if */
        else if ((blockAddr & 0xFC00) == 0x0C00) {
          linAddr = 0x13FC00 + (blockAddr & 0x03FF);
        } /* else if */
        else {
          continue; /* not a paged address, skip */
        } /* else */
        blockAddr = linAddr;
      } /* if */

      /* HC12: build linear Address, skip EEPROM data */
      else if (targetIsHC12 && pagedAddress) {
        if ((blockAddr & 0xC000) == 0x8000) {
          linAddr  = (blockAddr & 0xFFFF0000) >> 2;
          linAddr += (blockAddr & 0x3FFF);
          linAddr &= (memFlashSize - 1);
        } /* if */
        else if ((blockAddr & 0xC000) == 0xC000) {
          linAddr = memFlashSize - 0x4000 + (blockAddr & 0x3FFF);
        } /* else if */
        else if ((blockAddr & 0xC000) == 0x4000) {
          linAddr = memFlashSize - 0x8000 + (blockAddr & 0x3FFF);
        } /* else if */
        else {
          continue; /* not a paged address, skip */
        } /* else */
        blockAddr = linAddr;
      } /* if */

      if ((memWriteMode == BDI_HC12_FLASH) || (memWriteMode == BDI_HC12A_FLASH)) {
        if (    (cpuTypeHC12 == BDI_CPU_HC912DG128)  || (cpuTypeHC12 == BDI_CPU_HC912DA128)
             || (cpuTypeHC12 == BDI_CPU_HC912DG128A) || (cpuTypeHC12 == BDI_CPU_HC912DT128A)
           ) {
          /* if block crosses 16kB boundary, split it */
          pData = blockData;
          splitCount = 0x4000 - ((WORD)blockAddr & 0x3FFF);
          if (splitCount < blockCount) {
            pageAddr = ((blockAddr << 2) & 0x70000) | (blockAddr & 0x3FFF) | 0x8000;
            result = BDI_Download(pageAddr, splitCount, pData);
            blockCount -= splitCount;
            blockAddr  += splitCount;
            pData      += splitCount;
          } /* if */
          pageAddr = ((blockAddr << 2) & 0x70000) | (blockAddr & 0x3FFF) | 0x8000;
          result = BDI_Download(pageAddr, (WORD)blockCount, pData);
        } /* if */
        else {
          result = BDI_Download(blockAddr, (WORD)blockCount, blockData);
        } /* else */
      } /* else if */

      else if (targetIsHC12 && (memWriteMode == BDI_STAR12_FLASH)) {
        /* Motorol type of linear address, aligned to top of 1MB */
        if (blockAddr >= memFlashSize) {
          blockAddr -= (0x100000 - memFlashSize);
        } /* else if */
        /* if block crosses 16kB boundary, split it */
        pData = blockData;
        splitCount = 0x4000 - ((WORD)blockAddr & 0x3FFF);
        if (splitCount < blockCount) {
          pageAddr = ((blockAddr << 2) & 0x3F0000) | (blockAddr & 0x3FFF);
          if      (cpuTypeHC12 == BDI_CPU_S12_512K4_4K) pageAddr += 0x208000;
          else if (cpuTypeHC12 == BDI_CPU_S12_256K4_4K) pageAddr += 0x308000;
          else if (cpuTypeHC12 == BDI_CPU_S12_128K2_2K) pageAddr += 0x388000;
          else if (cpuTypeHC12 == BDI_CPU_S12_64K1_1K)  pageAddr += 0x3C8000;
          else if (cpuTypeHC12 == BDI_CPU_S12_32K1_1K)  pageAddr += 0x3E8000;
          else if (cpuTypeHC12 == BDI_CPU_S12_256K2_4K) pageAddr += 0x308000;
          else if (cpuTypeHC12 == BDI_CPU_S12_256K2_2K) pageAddr += 0x308000;
          else if (cpuTypeHC12 == BDI_CPU_S12_128K1_2K) pageAddr += 0x388000;
          result = BDI_Download(pageAddr, splitCount, pData);
          blockCount -= splitCount;
          blockAddr  += splitCount;
          pData      += splitCount;
        } /* if */
        pageAddr = ((blockAddr << 2) & 0x3F0000) | (blockAddr & 0x3FFF);
        if      (cpuTypeHC12 == BDI_CPU_S12_512K4_4K) pageAddr += 0x208000;
        else if (cpuTypeHC12 == BDI_CPU_S12_256K4_4K) pageAddr += 0x308000;
        else if (cpuTypeHC12 == BDI_CPU_S12_128K2_2K) pageAddr += 0x388000;
        else if (cpuTypeHC12 == BDI_CPU_S12_64K1_1K)  pageAddr += 0x3C8000;
        else if (cpuTypeHC12 == BDI_CPU_S12_32K1_1K)  pageAddr += 0x3E8000;
        else if (cpuTypeHC12 == BDI_CPU_S12_256K2_4K) pageAddr += 0x308000;
        else if (cpuTypeHC12 == BDI_CPU_S12_256K2_2K) pageAddr += 0x308000;
        else if (cpuTypeHC12 == BDI_CPU_S12_128K1_2K) pageAddr += 0x388000;
        result = BDI_Download(pageAddr, (WORD)blockCount, pData);
      } /* else if */

      else if (targetIsS12X && (memWriteMode == BDI_STAR12_FLASH)) {
        /* if block crosses 16kB boundary, split it */
        pData = blockData;
        splitCount = 0x4000 - ((WORD)blockAddr & 0x3FFF);
        if (splitCount < blockCount) {
          pageAddr = ((blockAddr << 2) & 0xFF0000) | ((blockAddr & 0x3FFF) + 0x8000);
          result = BDI_Download(pageAddr, splitCount, pData);
          blockCount -= splitCount;
          blockAddr  += splitCount;
          pData      += splitCount;
        } /* if */
        pageAddr = ((blockAddr << 2) & 0xFF0000) | ((blockAddr & 0x3FFF) + 0x8000);
        result = BDI_Download(pageAddr, (WORD)blockCount, pData);
      } /* else if */

      else if (targetIsS12X && (memWriteMode == BDI_STAR12_EEPROM)) {
        /* if block crosses 1kB boundary, split it */
        pData = blockData;
        splitCount = 0x0400 - ((WORD)blockAddr & 0x03FF);
        if (splitCount < blockCount) {
          pageAddr = ((blockAddr << 6) & 0xFF0000) | ((blockAddr & 0x03FF) + 0x0800);
          result = BDI_Download(pageAddr, splitCount, pData);
          blockCount -= splitCount;
          blockAddr  += splitCount;
          pData      += splitCount;
        } /* if */
        pageAddr = ((blockAddr << 6) & 0xFF0000) | ((blockAddr & 0x03FF) + 0x0800);
        result = BDI_Download(pageAddr, (WORD)blockCount, pData);
      } /* else if */

      else {
        /* fill 8-byte phrase with 0xff */
        if (memWriteMode == BDI_S12XE_PFLASH) {
          while (blockCount & 7) {
            blockData[blockCount] = 0xff;
            blockCount++;
          } /* while */
        } /* if */

        /* if block crosses 16kB boundary, split it */
        pData = blockData;
        splitCount = 0x4000 - ((WORD)blockAddr & 0x3FFF);
        if (splitCount < blockCount) {
          result = BDI_Download(blockAddr, splitCount, pData);
          blockCount -= splitCount;
          blockAddr  += splitCount;
          pData      += splitCount;
        } /* if */
        result = BDI_Download(blockAddr, (WORD)blockCount, pData);
      } /* else */

      /* progress callback */
      if (progressCallback != NULL) {
        if (progressCallback(totalCount) == BDI_ABORT_REQUEST) result = BDI_ERR_USER_ABORT;
      } /* if */

    } /* if */
    else if (blockCount < 0) {
      result = blockCount;
    } /* else if */

  } while ((result == BDI_OKAY) && (blockCount > 0));

  /* terminate download */
  if (result == BDI_OKAY) {
    result = BDI_EndDownload(errorAddr);
  } /* if */
  else {
    (void)BDI_EndDownload(errorAddr);
  } /* else */

  if (result == BDI_OKAY) {
    if (!feof(file)) result = BDI_ERR_FILE_ACCESS;
  } /* if */

  fclose(file);

  return result;
} /* BDI_WriteFile */


/****************************************************************************
 ****************************************************************************

    Writes a binary file into the target memory
    The Write mode of the BDI is already set.

 ****************************************************************************/

static int BDI_WriteBinary(const char *fileName, DWORD startAddr, DWORD *errorAddr)
{
  int       result;
  FILE*     file;
  BYTE      blockData[BDI_MAX_BLOCK_SIZE];
  int       blockCount;
  DWORD     totalCount;
  DWORD     blockAddr;

  /* open binary file */
  file = fopen(fileName,"rb");
  if (file == NULL) return BDI_ERR_FILE_ACCESS;

  /* init download */
  result = BDI_InitDownload(BDI_NO_VERIFY, TRUE);
  if (result != BDI_OKAY) {
    fclose(file);
    return result;
  } /* if */

  /* download data */
  totalCount = 0;
  blockAddr  = startAddr;
  do {
    blockCount = fread(blockData, 1, sizeof blockData, file);
    if (blockCount > 0) {
      result = BDI_Download(blockAddr, (WORD)blockCount, blockData);
      blockAddr  += blockCount;
      totalCount += blockCount;
    } /* if */

    /* progress callback */
    if (progressCallback != NULL) {
      if (progressCallback(totalCount) == BDI_ABORT_REQUEST) result = BDI_ERR_USER_ABORT;
    } /* if */

  } while ((result == BDI_OKAY) && (blockCount > 0));

  /* terminate download */
  if (result == BDI_OKAY) {
    result = BDI_EndDownload(errorAddr);
  } /* if */
  else {
    (void)BDI_EndDownload(errorAddr);
  } /* else */

  if (result == BDI_OKAY) {
    if (!feof(file)) result = BDI_ERR_FILE_ACCESS;
  } /* if */

  fclose(file);

  return result;
} /* BDI_WriteBinary */


/****************************************************************************
 ****************************************************************************

    Writes a block to the target RAM memory

    INPUT:  addr            address of the target
            count           number of bytes in the block (up to 64k)
            block           the data block
    OUTPUT: errorAddr       address of the failing byte
            return          error code

 ****************************************************************************/

int BDI_LoadBlock(DWORD addr, WORD count, BYTE *block, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (memWriteMode != BDI_LOAD_MODE_RAM)
  {
    memWriteMode = BDI_LOAD_MODE_RAM;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, memOption1);
  } /* if */

  /* write block */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteBlock(addr, count, block, errorAddr);
  }

  return result;
} /* BDI_LoadBlock */


/****************************************************************************
 ****************************************************************************

    Writes a S-Record file into the target RAM memory

    INPUT:  fileName        the full path and filename of the S-record file
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_LoadFile2(const char *fileName, DWORD offset, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (memWriteMode != BDI_LOAD_MODE_RAM)
  {
    memWriteMode = BDI_LOAD_MODE_RAM;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, memOption1);
  }

  /* write block */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteFile(fileName, offset, errorAddr);
  }

  return result;
} /* BDI_LoadFile2 */

int BDI_LoadFile(const char *fileName, DWORD *errorAddr)
{
  return BDI_LoadFile2(fileName, 0, errorAddr);
} /* BDI_LoadFile */


/****************************************************************************
 ****************************************************************************

    Writes a Binary file into the target RAM memory

 ****************************************************************************/

int BDI_LoadBinary(const char *fileName, DWORD startAddr, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (memWriteMode != BDI_LOAD_MODE_RAM)
  {
    memWriteMode = BDI_LOAD_MODE_RAM;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, memOption1);
  }

  /* write file */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteBinary(fileName, startAddr, errorAddr);
  }

  return result;
} /* BDI_LoadBinary */


/****************************************************************************
 ****************************************************************************

    Compares a data block against the target memory

    INPUT:  addr            address of the target memory
            count           number of bytes in the block (up to 64k)
            block           the data block
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_VerifyBlock(DWORD addr, WORD count, BYTE *block, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (memWriteMode != BDI_LOAD_MODE_VERIFY)
  {
    memWriteMode = BDI_LOAD_MODE_VERIFY;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, memOption1);
  }

  /* write block */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteBlock(addr, count, block, errorAddr);
  }

  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_VERIFY;
  return result;
} /* BDI_VerifyBlock */


/****************************************************************************
 ****************************************************************************

    Compares a S-Record file against the target memory

    INPUT:  fileName        the full path and filename of the S-record file
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_VerifyFile2(const char *fileName, DWORD offset, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (memWriteMode != BDI_LOAD_MODE_VERIFY)
  {
    memWriteMode = BDI_LOAD_MODE_VERIFY;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, memOption1);
  }

  /* write block */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteFile(fileName, offset, errorAddr);
  }

  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_VERIFY;
  return result;
} /* BDI_VerifyFile2 */

int BDI_VerifyFile(const char *fileName, DWORD *errorAddr)
{
  return BDI_VerifyFile2(fileName, 0, errorAddr);
} /* BDI_VerifyFile */


/****************************************************************************
 ****************************************************************************

    Compares a Binary file against the target memory

 ****************************************************************************/

int BDI_VerifyBinary(const char *fileName, DWORD startAddr, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (memWriteMode != BDI_LOAD_MODE_VERIFY)
  {
    memWriteMode = BDI_LOAD_MODE_VERIFY;
    result = BDI_SetWriteMode(memWriteMode, memTypeHC12, 0, 0, memOption1);
  }

  /* write file */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteBinary(fileName, startAddr, errorAddr);
  }

  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_VERIFY;
  return result;
} /* BDI_VerifyBinary */


/****************************************************************************
 ****************************************************************************

    Calculate the number of bytes in the file

    INPUT:  fileName        the full path and filename of the S-record file
    OUTPUT: return          count if okay or a negativ number if error

 ****************************************************************************/

int BDI_FileSize(const char *fileName)
{
  char      szLine[SREC_MAX_LINE];
  BYTE      data[SREC_MAX_DATA];
  int       count;
  int       totalCount;
  DWORD     addr;
  FILE*     file;

  /* open S-Record file */
  file = fopen(fileName,"r");
  if (file == NULL) return BDI_ERR_FILE_ACCESS;

  /* count the data */
  totalCount = 0;
  while (fgets(szLine, sizeof szLine, file) != NULL) {
    count = DecodeSRecord(szLine, &addr, data);
    if (count > 0) {
      totalCount += count;
    } /* if */
    else if (count < 0) {
      fclose(file);
      return BDI_ERR_FILE_ACCESS;
    } /* else if */
  } /* while */

  fclose(file);
  return totalCount;
} /* BDI_FileSize */


void BDI_InstallCallback(BDI_ProgressCallbackT* cb)
{
  progressCallback = cb;
} /* BDI_InstallCallback */


/****************************************************************************
 ****************************************************************************
                Low level JTAG access
 ****************************************************************************
 ****************************************************************************/

#define JTAG_SCAN_DROP          (1<<7)

int BDI_DoJtag(      BYTE  command,
                     BYTE  control,
                     WORD  count,
               const BYTE* pScanOut,
                     BYTE* pScanIn)
{
  BYTE*     cmdPtr;
  BYTE*     ansPtr;
  int       rxCount;
  BYTE      answer;
  WORD      errorCode;
  WORD      byteCount;
  WORD      i;

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, "BDI_DoJtag(%u, 0x%02x, %u)", command, control, count);
  } /* if */
#endif

  /* check parameters */
  if ((command == JTAG_CMD_SCAN) || (command == JTAG_CMD_IRSCAN) || (command == JTAG_CMD_DRSCAN)) {
    byteCount = (count + 7) / 8;
    if (pScanIn == NULL) control |= JTAG_SCAN_DROP;
  } /* if */
  else {
    byteCount = 0;
  } /* else */

  if (byteCount > BDI_MAX_BLOCK_SIZE) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_PARAMETER));
    } /* if */
#endif
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_CMD_JTAG_IO, cmdBuffer);
  cmdPtr = BDI_AppendByte(command,         cmdPtr);
  cmdPtr = BDI_AppendByte(control,         cmdPtr);
  cmdPtr = BDI_AppendWord(count,           cmdPtr);
  if ((pScanOut != NULL) && byteCount) {
    for (i = 0; i < byteCount; i++) *cmdPtr++ = *pScanOut++;
  } /* if */

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer,
                            cmdBuffer,
                            sizeof ansBuffer,
                            ansBuffer,
                            200 + byteCount);
  if (rxCount < 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(rxCount));
    } /* if */
#endif
    return rxCount;
  } /* if */

  /* analyse response */
  ansPtr = ansBuffer;
  answer = *ansPtr++;
  if (answer != BDI_CMD_JTAG_IO) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet(BDI_ERR_INVALID_RESPONSE));
    } /* if */
#endif
    return BDI_ERR_INVALID_RESPONSE;
  } /* if */

  /* get error code */
  ansPtr = BDI_ExtractWord(&errorCode, ansPtr);
  if (errorCode != 0) {
#ifdef _DEBUG
    if (logFile != NULL) {
      fprintf(logFile, " => FAILED (%s)\n", BDI_ErrorTextGet((int)(short)errorCode));
    } /* if */
#endif
    return (int)(short)errorCode;
  } /* if */

  /* get read data */
  if (byteCount && (pScanIn != NULL)) {
    while (byteCount--) *pScanIn++ = *ansPtr++;
  } /* if */

#ifdef _DEBUG
  if (logFile != NULL) {
    fprintf(logFile, " => OKAY\n");
  } /* if */
#endif
  return BDI_OKAY;
} /* BDI_DoJtag */


/****************************************************************************
 ****************************************************************************
                Flash Programming Functions
 ****************************************************************************
 ****************************************************************************/

/* erase state codes */
#define BDI_ERASE_TERMINATED    0   /* successfully terminated  */
#define BDI_ERASE_PROGRAM       1   /* program to 0 in process  */
#define BDI_ERASE_ERASE         2   /* erase/verify  in process */
#define BDI_ERASE_FAILED        3   /* erase failed (error at actual address) */


/****************************************************************************
 ****************************************************************************

 Start erase of a bank of the target memory.

  INPUT:  eraseType       type of erase algorithm
          busWidth        Targe memory bus width
          chipSize        size of the chip or segment to erase
          baseAddr        base address of the memory bank
          ramBase         base address of ram in the target system for
                          the algorithm code
          option1         an optional parameter, depends on erase type
  OUTPUT: return          error code

 ****************************************************************************/

static int BDI_StartErase(BYTE   eraseType,
                          BYTE   busWidth,
                          DWORD  chipSize,
                          DWORD  baseAddr,
                          DWORD  ramBase,
                          DWORD  option1)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;
  BYTE      result;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_PRO_START_ERASE, cmdBuffer);
  cmdPtr = BDI_AppendByte(eraseType, cmdPtr);
  cmdPtr = BDI_AppendByte(busWidth,  cmdPtr);
  cmdPtr = BDI_AppendLong(chipSize,  cmdPtr);
  cmdPtr = BDI_AppendLong(baseAddr,  cmdPtr);
  cmdPtr = BDI_AppendLong(ramBase,   cmdPtr);
  cmdPtr = BDI_AppendLong(option1,   cmdPtr);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer, cmdBuffer, sizeof ansBuffer, ansBuffer, 5000);
  if (rxCount < 0) return rxCount;

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 2) || (answer != BDI_PRO_START_ERASE))
    return BDI_ERR_INVALID_RESPONSE;

  ansPtr = BDI_ExtractByte(&result, ansPtr);
  if      (result == 0) return BDI_OKAY;
  else if (result == 1) return BDI_ERR_INVALID_MODE;
  else if (result == 3) return BDI_ERR_WORKING_RAM;
  else if (result == 4) return BDI_ERR_INVALID_PARAMETER;
  else                  return BDI_ERR_INVALID_RESPONSE;
} /* BDI_StartErase */


/****************************************************************************
 ****************************************************************************

 Get state of the erase process.

  INPUT:  -
  OUTPUT: eraseState      state of the erase process
          actualAddr      actual address in process or
                          the address of the failing byte
          return          error code (communication errors)

 ****************************************************************************/

static int BDI_GetEraseState(BYTE *eraseState, DWORD *actualAddr)
{
  BYTE     *cmdPtr;
  BYTE     *ansPtr;
  int       rxCount;
  BYTE      answer;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_PRO_GET_ERASE_STATE, cmdBuffer);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer, cmdBuffer, sizeof ansBuffer, ansBuffer, 5000);
  if (rxCount < 0) return rxCount;

  /* analyse response */
  ansPtr = BDI_ExtractByte(&answer, ansBuffer);
  if ((rxCount != 6) || (answer != BDI_PRO_GET_ERASE_STATE))
    return BDI_ERR_INVALID_RESPONSE;

  ansPtr = BDI_ExtractByte(eraseState, ansPtr);
  ansPtr = BDI_ExtractLong(actualAddr, ansPtr);
  return BDI_OKAY;
} /* BDI_GetEraseState */


/****************************************************************************
 ****************************************************************************

    Select bdiPro mode

    INPUT:  -
    OUTPUT: return          error code

 ****************************************************************************/


static int BDI_SetProMode(void)
{
  BYTE     *cmdPtr;
  int       rxCount;

  /* prepare command */
  cmdPtr = BDI_AppendByte(BDI_PRO_SET_PRO_MODE, cmdBuffer);

  /* BDI transaction */
  rxCount = BDI_Transaction(cmdPtr-cmdBuffer, cmdBuffer, sizeof ansBuffer, ansBuffer, 2000);
  if (rxCount < 0) return rxCount;

  return BDI_OKAY;
} /* BDI_SetProMode */


/****************************************************************************
 ****************************************************************************

    Defines the  Flash memory

    INPUT:  flashType       the used Flash Memory Type
            flashSize       the size of one flash memory chip in bytes
            pageSize        the size in bytes of a page of one chip
                            enter 0 if the chip has no pages
            memoryWidth     the memory bus width in bits (8,16,32)
            workSpace       some programming algorithm must run in the target
                            CPU, enter the base address of a RAM area reseved
                            for this algorithm.
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashSetType(WORD  flashType,
                     DWORD flashSize,
                     WORD  pageSize,
                     WORD  memoryWidth,
                     DWORD workSpace)
{

  if (    (flashType != BDI_FLASH_AM29F0X0)
       && (flashType != BDI_FLASH_I28FBXB)
       && (flashType != BDI_FLASH_I28FBXW)
       && (flashType != BDI_FLASH_AM29FX00B)
       && (flashType != BDI_FLASH_AM29FX00W)
       && (flashType != BDI_FLASH_AT49)
       && (flashType != BDI_FLASH_AT49X8)
       && (flashType != BDI_FLASH_AT49X16)
       && (flashType != BDI_FLASH_STRATAX8)
       && (flashType != BDI_FLASH_STRATAX16)
       && (flashType != BDI_FLASH_MIRROR)
       && (flashType != BDI_FLASH_MIRRORX8)
       && (flashType != BDI_FLASH_MIRRORX16)
       && (flashType != BDI_FLASH_S29M64X8)
       && (flashType != BDI_FLASH_S29M32X16)
       && (flashType != BDI_FLASH_I28FBXD)
       && (flashType != BDI_FLASH_AM29DX16)
       && (flashType != BDI_FLASH_AM29DX32)
       && (flashType != BDI_FLASH_AM29BDDX16)
       && (flashType != BDI_FLASH_AM29BDDX32)
       && (flashType != BDI_FLASH_CFM)
       && (flashType != BDI_FLASH_CFM16)
       && (flashType != BDI_FLASH_CFM32)
       && (flashType != BDI_FLASH_STA2051)
       && (flashType != BDI_FLASH_ADUC7000)
       && (flashType != BDI_FLASH_ST30F)
       && (flashType != BDI_FLASH_AT91SAM7S)
       && (flashType != BDI_FLASH_STR910F)
       && (flashType != BDI_FLASH_H7F)
       && (flashType != BDI_FLASH_STM32F10)
       && (flashType != BDI_FLASH_LM3S)
       && (flashType != BDI_FLASH_SAM9XE)
       && (flashType != BDI_FLASH_SAM7L)
       && (flashType != BDI_FLASH_STM32L15)
       && (flashType != BDI_FLASH_EFM32)
       && (flashType != BDI_FLASH_FTFL)
       && (flashType != BDI_FLASH_FTFE)
       && (flashType != BDI_FLASH_STM32F2)
       && (flashType != BDI_FLASH_STM32F4)
       && (flashType != BDI_FLASH_STM32W)
       && (flashType != BDI_FLASH_S29GLSX16)
       && (flashType != BDI_FLASH_S29WSRX16)
       && (flashType != BDI_FLASH_S29VSRX16)
     )
  {
    return BDI_ERR_INVALID_PARAMETER;
  }
  memFlashType = (BYTE)flashType;
  flashChanged = TRUE;

  if (    (flashType == BDI_FLASH_I28FBXW)
       || (flashType == BDI_FLASH_AM29FX00W)
       || (flashType == BDI_FLASH_AT49X16)
       || (flashType == BDI_FLASH_STRATAX16)
       || (flashType == BDI_FLASH_MIRRORX16)
       || (flashType == BDI_FLASH_S29M32X16)
       || (flashType == BDI_FLASH_AM29DX16)
       || (flashType == BDI_FLASH_AM29BDDX16)
       || (flashType == BDI_FLASH_CFM16)
       || (flashType == BDI_FLASH_ADUC7000)
       || (flashType == BDI_FLASH_STR910F)
       || (flashType == BDI_FLASH_STM32F10)
       || (flashType == BDI_FLASH_S29GLSX16)
       || (flashType == BDI_FLASH_S29WSRX16)
       || (flashType == BDI_FLASH_S29VSRX16)
       ) {
    memFlashWidth = 2;
  } /* if */
  else if (    (flashType == BDI_FLASH_I28FBXD)
            || (flashType == BDI_FLASH_AM29DX32)
            || (flashType == BDI_FLASH_AM29BDDX32)
            || (flashType == BDI_FLASH_CFM)
            || (flashType == BDI_FLASH_CFM32)
            || (flashType == BDI_FLASH_STA2051)
            || (flashType == BDI_FLASH_ST30F)
            || (flashType == BDI_FLASH_AT91SAM7S)
            || (flashType == BDI_FLASH_H7F)
            || (flashType == BDI_FLASH_LM3S)
            || (flashType == BDI_FLASH_SAM9XE)
            || (flashType == BDI_FLASH_SAM7L)
            || (flashType == BDI_FLASH_STM32L15)
            || (flashType == BDI_FLASH_EFM32)
            || (flashType == BDI_FLASH_FTFL)
            || (flashType == BDI_FLASH_FTFE)
          ) {
    memFlashWidth = 4;
  } /* if */
  else {
    memFlashWidth = 1;
  } /* else */

  if      (memoryWidth == 8)  memBusWidth = 1;
  else if (memoryWidth == 16) memBusWidth = 2;
  else if (memoryWidth == 32) memBusWidth = 4;
  else if (memoryWidth == 64) memBusWidth = 8;
  else                        return BDI_ERR_INVALID_PARAMETER;

  memPageSize  = (DWORD)pageSize * (DWORD)memBusWidth / (DWORD)memFlashWidth;
  memRamBase   = workSpace;
  memFlashSize = flashSize;
  if (flashType == BDI_FLASH_AT91SAM7S) {
    memBankMask = flashSize;
  } /* if */
  else {
    memBankMask = 0xFFFFFFFF - ( (flashSize * (DWORD)memBusWidth / (DWORD)memFlashWidth) - 1);
  } /* else */
  if (flashType == BDI_FLASH_H7F) {
    memOption1 = (DWORD)pageSize;
  } /* if */

  return BDI_OKAY;
} /* BDI_FlashSetType */


/****************************************************************************
 ****************************************************************************

    Programs a data block into the target flash memory

    INPUT:  addr            address of the target memory
            count           number of bytes in the block (up to 64k)
            block           the data block
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashWriteBlock(DWORD addr, WORD count, BYTE *block, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (flashChanged || (memWriteMode != memFlashType))
  {
    memWriteMode = memFlashType;
    result = BDI_SetWriteMode(memWriteMode, memBusWidth, memBankMask, memRamBase, memOption1);
    flashChanged = FALSE;
  }

  /* write block */
  if (result == BDI_OKAY) {
    result = BDI_WriteBlock(addr, count, block, errorAddr);
  } /* if */

  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_FLASH_PROGRAM;
  return result;
} /* BDI_FlashWriteBlock */


/****************************************************************************
 ****************************************************************************

    Programs a S-Record file into the target flash memory

    INPUT:  fileName        the full path and filename of the S-record file
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashWriteFile2(const char *fileName, DWORD offset, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (flashChanged || (memWriteMode != memFlashType))
  {
    memWriteMode = memFlashType;
    result = BDI_SetWriteMode(memWriteMode, memBusWidth, memBankMask, memRamBase, memOption1);
    flashChanged = FALSE;
  }

  /* write block */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteFile(fileName, offset, errorAddr);
  }

  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_FLASH_PROGRAM;
  return result;
} /* BDI_FlashWriteFile2 */

int BDI_FlashWriteFile(const char *fileName, DWORD *errorAddr)
{
  return BDI_FlashWriteFile2(fileName, 0, errorAddr);
} /* BDI_FlashWriteFile */


/****************************************************************************
 ****************************************************************************

    Programs a Binary file into the target flash memory

 ****************************************************************************/

int BDI_FlashWriteBinary(const char *fileName, DWORD startAddr, DWORD *errorAddr)
{
  int   result = BDI_OKAY;

  /* set write mode */
  if (flashChanged || (memWriteMode != memFlashType))
  {
    memWriteMode = memFlashType;
    result = BDI_SetWriteMode(memWriteMode, memBusWidth, memBankMask, memRamBase, memOption1);
    flashChanged = FALSE;
  }

  /* write file */
  if (result == BDI_OKAY)
  {
    result = BDI_WriteBinary(fileName, startAddr, errorAddr);
  }

  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_FLASH_PROGRAM;
  return result;
} /* BDI_FlashWriteBinary */


/****************************************************************************
 ****************************************************************************
                Flash Erase Functions
 ****************************************************************************
 ****************************************************************************/

#define ERASE_TIMEOUT             60000     /* Erase timeout (ms) */

#define AMD_EXCEEDED_TIME_LIMIT   0x20


/****************************************************************************
 ****************************************************************************

    Erase Am29F010 family

    INPUT:  mode            Erase mode
            sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseAm29F010(WORD mode, DWORD sectorAddr)
{
  int       result;
  DWORD     dwBaseAddr;
  BYTE      oldByte;
  BYTE      newByte;
  WORD      oldWord;
  WORD      newWord;
  DWORD     oldLong;
  DWORD     newLong;
  BYTE      oldDL[8];
  BYTE      newDL[8];

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  static const BYTE PatternA[]  = {0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA};
  static const BYTE PatternB[]  = {0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55};
  static const BYTE EraseCmd[]  = {0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80};
  static const BYTE ChipCmd[]   = {0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10};
  static const BYTE SectorCmd[] = {0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30};
  static const BYTE ResetCmd[]  = {0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0};
  static const BYTE Erased[]    = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

  result     = BDI_OKAY;
  dwBaseAddr = sectorAddr & memBankMask;
  startTime  = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth)
  {
    case 1: /* 8 BIT */
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x002AAA, 0x55);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0x80);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x002AAA, 0x55);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0x10);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0x30);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetByte(sectorAddr, &oldByte);
      while ((result == BDI_OKAY) && (oldByte != 0xFF)) {
        result = BDI_GetByte(sectorAddr, &newByte);
        if (result == BDI_OKAY) {
          if ((newByte != 0xFF) && ((oldByte & AMD_EXCEEDED_TIME_LIMIT) != 0)) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldByte = newByte;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetByte(dwBaseAddr + 0x005555, 0xAA);
      (void)BDI_SetByte(dwBaseAddr + 0x002AAA, 0x55);
      (void)BDI_SetByte(dwBaseAddr + 0x005555, 0xF0);
      break;


    case 2: /* 16 BIT */
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x005554, 0x5555);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x8080);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x005554, 0x5555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x1010);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &oldWord);
      while ((result == BDI_OKAY) && (oldWord != 0xFFFF)) {
        result = BDI_GetWord(sectorAddr, &newWord);
        if (result == BDI_OKAY) {
          if (   (((newWord & 0xFF00) != 0xFF00) && (((oldWord>>8) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newWord & 0x00FF) != 0x00FF) && (((oldWord>>0) & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldWord = newWord;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xAAAA);
      (void)BDI_SetWord(dwBaseAddr + 0x005554, 0x5555);
      (void)BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xF0F0);
      break;


    case 4: /* 32BIT */
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x10101010);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &oldLong);
      while ((result == BDI_OKAY) && (oldLong != 0xFFFFFFFF)) {
        result = BDI_GetLong(sectorAddr, &newLong);
        if (result == BDI_OKAY) {
          if (   (((newLong & 0xFF000000) != 0xFF000000) && (((oldLong>>24) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x00FF0000) != 0x00FF0000) && (((oldLong>>16) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x0000FF00) != 0x0000FF00) && (((oldLong>>8)  & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x000000FF) != 0x000000FF) && (((oldLong>>0)  & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldLong = newLong;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      (void)BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      (void)BDI_SetLong(dwBaseAddr + 0x015554, 0xF0F0F0F0);
      break;

    case 8: /* 64BIT */
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, EraseCmd);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, ChipCmd);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, oldDL);
      while ((result == BDI_OKAY) && (memcmp(oldDL, Erased, 8) != 0)) {
        result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, newDL);
        if (result == BDI_OKAY) {
          if (   ((newDL[0] != 0xFF) && ((oldDL[0] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[1] != 0xFF) && ((oldDL[1] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[2] != 0xFF) && ((oldDL[2] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[3] != 0xFF) && ((oldDL[3] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[4] != 0xFF) && ((oldDL[4] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[5] != 0xFF) && ((oldDL[5] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[6] != 0xFF) && ((oldDL[6] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[7] != 0xFF) && ((oldDL[7] & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            memcpy(oldDL, newDL, sizeof oldDL);
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      (void)BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      (void)BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, ResetCmd);
      break;

  } /* switch */

  return result;
} /* EraseAm29F010 */


/****************************************************************************
 ****************************************************************************

    Erase I28FBX family

    INPUT:  sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseI28FBX(DWORD sectorAddr)
{
  int       result;
  DWORD     dwBaseAddr;
  BYTE      stateByte;
  WORD      stateWord;
  DWORD     stateLong;
  WORD      wordMask;
  DWORD     longMask;

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  result     = BDI_OKAY;
  dwBaseAddr = sectorAddr & memBankMask;
  startTime  = times(&tbuf);

  switch (memBusWidth)
  {
    case 1: /* 8 BIT */
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0x20);
      if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0xD0);

      /* wait for end of erasing */
      while (result == BDI_OKAY)
      {
        if (result == BDI_OKAY) result = BDI_GetByte(sectorAddr, &stateByte);
        if (result == BDI_OKAY)
        {
          /* check for finished */
          if ((stateByte & 0x80) == 0x80)
          {
            if ((stateByte & 0x78) != 0)
              result = BDI_ERR_FLASH_ERASE;
            else
              break;
          } /* if */
          /* check for timeout */
          else if ((times(&tbuf) - startTime) > eraseTimeout)
          {
            result = BDI_ERR_FLASH_ERASE;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      BDI_SetByte(sectorAddr, 0x50);
      BDI_SetByte(sectorAddr, 0xFF);
      BDI_GetByte(sectorAddr, &stateByte);
      break;


    case 2: /* 16 BIT */
      /* start erasing */
      if (memFlashWidth == 2) wordMask = 0x00FF;
      else                    wordMask = 0xFFFF;
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, (WORD)(0x2020 & wordMask));
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, (WORD)(0xD0D0 & wordMask));

      /* wait for end of erasing */
      while (result == BDI_OKAY)
      {
        if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &stateWord);
        if (result == BDI_OKAY)
        {
          /* check for finished */
          if ((stateWord & 0x8080 & wordMask) == (0x8080 & wordMask))
          {
            if ((stateWord & 0x7878 & wordMask) != 0)
              result = BDI_ERR_FLASH_ERASE;
            else
              break;
          } /* if */
          /* check for timeout */
          else if ((times(&tbuf) - startTime) > eraseTimeout)
          {
            result = BDI_ERR_FLASH_ERASE;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      BDI_SetWord(sectorAddr, (WORD)(0x5050 & wordMask));
      BDI_SetWord(sectorAddr, (WORD)(0xFFFF & wordMask));
      BDI_GetWord(sectorAddr, &stateWord);
      break;


    case 4: /* 32BIT */
      /* start erasing */
      if      (memFlashWidth == 2) longMask = 0x00FF00FF;
      else if (memFlashWidth == 4) longMask = 0x000000FF;
      else                         longMask = 0xFFFFFFFF;
      if (memFlashWidth == 4) {  /* for ST / Numonix M58BW32F */
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x154, 0x20);
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0xD0);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x20202020 & longMask);
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0xD0D0D0D0 & longMask);
      } /* else */

      /* wait for end of erasing */
      while (result == BDI_OKAY)
      {
        if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &stateLong);
        if (result == BDI_OKAY)
        {
          /* check for finished */
          if ((stateLong & 0x80808080 & longMask) == (0x80808080 & longMask))
          {
            if ((stateLong & 0x78787878 & longMask) != 0)
              result = BDI_ERR_FLASH_ERASE;
            else
              break;
          } /* if */
          /* check for timeout */
          else if ((times(&tbuf) - startTime) > eraseTimeout)
          {
            result = BDI_ERR_FLASH_ERASE;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      BDI_SetLong(sectorAddr, 0x50505050 & longMask);
      BDI_SetLong(sectorAddr, 0xFFFFFFFF & longMask);
      BDI_GetLong(sectorAddr, &stateLong);
      break;

  } /* switch */

  return result;
} /* EraseI28FBX */


/****************************************************************************
 ****************************************************************************

    Erase Am29F100 family

    INPUT:  mirrorV2        Set if for Mirror V2 offsets
            mode            Erase mode
            sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseAm29F100(BOOL mirrorV2, WORD mode, DWORD sectorAddr)
{
  int       result;
  DWORD     dwBaseAddr;
  DWORD     cmdOffsetA;
  DWORD     cmdOffsetB;
  BYTE      oldByte;
  BYTE      newByte;
  WORD      oldWord;
  WORD      newWord;
  DWORD     oldLong;
  DWORD     newLong;
  BYTE      oldDL[8];
  BYTE      newDL[8];

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  static const BYTE PatternA[]  = {0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA};
  static const BYTE PatternB[]  = {0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55};
  static const BYTE EraseCmd[]  = {0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80};
  static const BYTE ChipCmd[]   = {0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10};
  static const BYTE SectorCmd[] = {0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30};
  static const BYTE ResetCmd[]  = {0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0};
  static const BYTE Erased[]    = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

  result      = BDI_OKAY;
  dwBaseAddr  = sectorAddr & memBankMask;
  startTime   = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth)
  {
    case 1: /* 8 BIT */
      /* start erasing */
      cmdOffsetA = 0x00AAAA;
      cmdOffsetB = 0x005555;
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + cmdOffsetA, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + cmdOffsetB, 0x55);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + cmdOffsetA, 0x80);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + cmdOffsetA, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + cmdOffsetB, 0x55);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + cmdOffsetA, 0x10);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0x30);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetByte(sectorAddr, &oldByte);
      while ((result == BDI_OKAY) && (oldByte != 0xFF)) {
        result = BDI_GetByte(sectorAddr, &newByte);
        if (result == BDI_OKAY) {
          if ((newByte != 0xFF) && ((oldByte & AMD_EXCEEDED_TIME_LIMIT) != 0)) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldByte = newByte;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetByte(dwBaseAddr + cmdOffsetA, 0xAA);
      (void)BDI_SetByte(dwBaseAddr + cmdOffsetB, 0x55);
      (void)BDI_SetByte(dwBaseAddr + cmdOffsetA, 0xF0);
      break;


    case 2: /* 16 BIT */
    if (memFlashWidth == 1)
    {
      /* start erasing */
      cmdOffsetA = 0x015554;
      cmdOffsetB = 0x00AAAA;
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetB, 0x5555);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0x8080);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetB, 0x5555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0x1010);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &oldWord);
      while ((result == BDI_OKAY) && (oldWord != 0xFFFF)) {
        result = BDI_GetWord(sectorAddr, &newWord);
        if (result == BDI_OKAY) {
          if (   (((newWord & 0xFF00) != 0xFF00) && (((oldWord>>8) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newWord & 0x00FF) != 0x00FF) && (((oldWord>>0) & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldWord = newWord;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xAAAA);
      (void)BDI_SetWord(dwBaseAddr + cmdOffsetB, 0x5555);
      (void)BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xF0F0);
    }
    else if (memFlashWidth == 2)
    {
      /* start erasing */
      cmdOffsetA = (mirrorV2) ? 0x0AAA : 0xAAAA;
      cmdOffsetB = (mirrorV2) ? 0x0554 : 0x5554;
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetB, 0x5555);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0x8080);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetB, 0x5555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + cmdOffsetA, 0x1010);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &oldWord);
      while ((result == BDI_OKAY) && (oldWord != 0xFFFF)) {
        result = BDI_GetWord(sectorAddr, &newWord);
        if (result == BDI_OKAY) {
          if ((newWord != 0xFFFF) && ((oldWord & AMD_EXCEEDED_TIME_LIMIT) != 0)) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldWord = newWord;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xAAAA);
      (void)BDI_SetWord(dwBaseAddr + cmdOffsetB, 0x5555);
      (void)BDI_SetWord(dwBaseAddr + cmdOffsetA, 0xF0F0);
    }
    break;


    case 4: /* 32BIT */
    if (memFlashWidth == 1)
    {
      /* start erasing */
      cmdOffsetA = 0x02AAA8;
      cmdOffsetB = 0x015554;
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetB, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetB, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0x10101010);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &oldLong);
      while ((result == BDI_OKAY) && (oldLong != 0xFFFFFFFF)) {
        result = BDI_GetLong(sectorAddr, &newLong);
        if (result == BDI_OKAY) {
          if (   (((newLong & 0xFF000000) != 0xFF000000) && (((oldLong>>24) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x00FF0000) != 0x00FF0000) && (((oldLong>>16) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x0000FF00) != 0x0000FF00) && (((oldLong>>8)  & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x000000FF) != 0x000000FF) && (((oldLong>>0)  & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldLong = newLong;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xAAAAAAAA);
      (void)BDI_SetLong(dwBaseAddr + cmdOffsetB, 0x55555555);
      (void)BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xF0F0F0F0);
    }
    else if (memFlashWidth == 2)
    {
      /* start erasing */
      cmdOffsetA = (mirrorV2) ? 0x01554 : 0x15554;
      cmdOffsetB = (mirrorV2) ? 0x00AA8 : 0x0AAA8;
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetB, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetB, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + cmdOffsetA, 0x10101010);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &oldLong);
      while ((result == BDI_OKAY) && (oldLong != 0xFFFFFFFF)) {
        result = BDI_GetLong(sectorAddr, &newLong);
        if (result == BDI_OKAY) {
          if (   (((newLong & 0xFFFF0000) != 0xFFFF0000) && (((oldLong>>16) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x0000FFFF) != 0x0000FFFF) && (((oldLong>>0)  & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldLong = newLong;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xAAAAAAAA);
      (void)BDI_SetLong(dwBaseAddr + cmdOffsetB, 0x55555555);
      (void)BDI_SetLong(dwBaseAddr + cmdOffsetA, 0xF0F0F0F0);
    }
    break;

    case 8: /* 64BIT */
    if (memFlashWidth == 1)
    {
      /* start erasing */
      cmdOffsetA = 0x55550;
      cmdOffsetB = 0x2AAA8;
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetB, BDI_AT_DEFAULT, 8, PatternB);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, EraseCmd);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetB, BDI_AT_DEFAULT, 8, PatternB);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, ChipCmd);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, oldDL);
      while ((result == BDI_OKAY) && (memcmp(oldDL, Erased, 8) != 0)) {
        result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, newDL);
        if (result == BDI_OKAY) {
          if (   ((newDL[0] != 0xFF) && ((oldDL[0] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[1] != 0xFF) && ((oldDL[1] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[2] != 0xFF) && ((oldDL[2] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[3] != 0xFF) && ((oldDL[3] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[4] != 0xFF) && ((oldDL[4] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[5] != 0xFF) && ((oldDL[5] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[6] != 0xFF) && ((oldDL[6] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || ((newDL[7] != 0xFF) && ((oldDL[7] & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            memcpy(oldDL, newDL, sizeof oldDL);
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, PatternA);
      (void)BDI_MemoryWrite(dwBaseAddr + cmdOffsetB, BDI_AT_DEFAULT, 8, PatternB);
      (void)BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, ResetCmd);
    }
    else if (memFlashWidth == 2)
    {
      /* start erasing */
      cmdOffsetA = (mirrorV2) ? 0x02AA8 : 0x2AAA8;
      cmdOffsetB = (mirrorV2) ? 0x01550 : 0x15550;
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetB, BDI_AT_DEFAULT, 8, PatternB);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, EraseCmd);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetB, BDI_AT_DEFAULT, 8, PatternB);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, ChipCmd);
      } /* if */
      else {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
      } /* else */

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, oldDL);
      while ((result == BDI_OKAY) && (memcmp(oldDL, Erased, 8) != 0)) {
        result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, newDL);
        if (result == BDI_OKAY) {
          if (   (((newDL[0] & newDL[1]) != 0xFF) && ((oldDL[1] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newDL[2] & newDL[3]) != 0xFF) && ((oldDL[3] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newDL[4] & newDL[5]) != 0xFF) && ((oldDL[5] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newDL[6] & newDL[7]) != 0xFF) && ((oldDL[7] & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            memcpy(oldDL, newDL, sizeof oldDL);
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, PatternA);
      (void)BDI_MemoryWrite(dwBaseAddr + cmdOffsetB, BDI_AT_DEFAULT, 8, PatternB);
      (void)BDI_MemoryWrite(dwBaseAddr + cmdOffsetA, BDI_AT_DEFAULT, 8, ResetCmd);
    }
    break;
  } /* switch */

  return result;
} /* EraseAm29F100 */


/****************************************************************************
 ****************************************************************************

    Erase AMD x32/x16 family

    INPUT:  mode            Erase mode
            sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseAM29D(WORD mode, DWORD sectorAddr)
{
  int       result;
  DWORD     dwBaseAddr;
  WORD      oldWord;
  WORD      newWord;
  DWORD     oldLong;
  DWORD     newLong;
  BYTE      oldDL[8];
  BYTE      newDL[8];

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  static const BYTE PatternA[]  = {0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA};
  static const BYTE PatternB[]  = {0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55};
  static const BYTE EraseCmd[]  = {0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80};
  static const BYTE ChipCmd[]   = {0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10};
  static const BYTE SectorCmd[] = {0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30};
  static const BYTE ResetCmd[]  = {0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0};
  static const BYTE Erased[]    = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

  result     = BDI_OKAY;
  dwBaseAddr = sectorAddr & memBankMask;
  startTime  = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth)
  {
    case 2: /* 16 BIT */
    if (memFlashWidth == 2)
    {
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x0AAAA, 0x5555);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0x8080);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x0AAAA, 0x5555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0x1010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
      }

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &oldWord);
      while ((result == BDI_OKAY) && (oldWord != 0xFFFF)) {
        result = BDI_GetWord(sectorAddr, &newWord);
        if (result == BDI_OKAY) {
          if ((newWord != 0xFFFF) && ((oldWord & AMD_EXCEEDED_TIME_LIMIT) != 0)) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldWord = newWord;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetWord(dwBaseAddr + 0x15554, 0xAAAA);
      (void)BDI_SetWord(dwBaseAddr + 0x0AAAA, 0x5555);
      (void)BDI_SetWord(dwBaseAddr + 0x15554, 0xF0F0);
    }
    break;


    case 4: /* 32BIT */
    if (memFlashWidth == 2)
    {
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x15554, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x15554, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0x10101010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      }

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &oldLong);
      while ((result == BDI_OKAY) && (oldLong != 0xFFFFFFFF)) {
        result = BDI_GetLong(sectorAddr, &newLong);
        if (result == BDI_OKAY) {
          if (   (((newLong & 0xFFFF0000) != 0xFFFF0000) && (((oldLong>>16) & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newLong & 0x0000FFFF) != 0x0000FFFF) && (((oldLong>>0)  & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldLong = newLong;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xAAAAAAAA);
      (void)BDI_SetLong(dwBaseAddr + 0x15554, 0x55555555);
      (void)BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xF0F0F0F0);
    }
    else if (memFlashWidth == 4)
    {
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x10101010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      }

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &oldLong);
      while ((result == BDI_OKAY) && (oldLong != 0xFFFFFFFF)) {
        result = BDI_GetLong(sectorAddr, &newLong);
        if (result == BDI_OKAY) {
          if ((newLong != 0xFFFFFFFF) && ((oldLong & AMD_EXCEEDED_TIME_LIMIT) != 0)) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            oldLong = newLong;
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      (void)BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      (void)BDI_SetLong(dwBaseAddr + 0x015554, 0xF0F0F0F0);
    }
    break;

    case 8: /* 64BIT */
    if (memFlashWidth == 2)
    {
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternB);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, EraseCmd);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternB);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, ChipCmd);
      }
      else {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
      }

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, oldDL);
      while ((result == BDI_OKAY) && (memcmp(oldDL, Erased, 8) != 0)) {
        result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, newDL);
        if (result == BDI_OKAY) {
          if (   (((newDL[0] & newDL[1]) != 0xFF) && ((oldDL[1] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newDL[2] & newDL[3]) != 0xFF) && ((oldDL[3] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newDL[4] & newDL[5]) != 0xFF) && ((oldDL[5] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newDL[6] & newDL[7]) != 0xFF) && ((oldDL[7] & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            memcpy(oldDL, newDL, sizeof oldDL);
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, PatternA);
      (void)BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternB);
      (void)BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, ResetCmd);
    }
    else if (memFlashWidth == 4)
    {
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, EraseCmd);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, ChipCmd);
      }
      else {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
      }

      /* wait for end of erasing */
      if (result == BDI_OKAY) result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, oldDL);
      while ((result == BDI_OKAY) && (memcmp(oldDL, Erased, 8) != 0)) {
        result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, newDL);
        if (result == BDI_OKAY) {
          if (   (((newDL[0] & newDL[1] & newDL[2] & newDL[3]) != 0xFF) && ((oldDL[3] & AMD_EXCEEDED_TIME_LIMIT) != 0))
              || (((newDL[4] & newDL[5] & newDL[6] & newDL[7]) != 0xFF) && ((oldDL[7] & AMD_EXCEEDED_TIME_LIMIT) != 0))
             ) {
            result = BDI_ERR_FLASH_ERASE;
          } /* if */
          else {
            if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
            memcpy(oldDL, newDL, sizeof oldDL);
          } /* else */
        } /* if */
      } /* while */

      /* Reset chip */
      (void)BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      (void)BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      (void)BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, ResetCmd);
    }
    break;
  } /* switch */

  return result;
} /* EraseAM29D */


static int EraseAM29BDD(WORD mode, DWORD sectorAddr)
{
  int       result;
  DWORD     dwBaseAddr;
  WORD      wValue;
  DWORD     lValue;

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  result     = BDI_OKAY;
  dwBaseAddr = sectorAddr & memBankMask;
  startTime  = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth)
  {
    case 2: /* 16 BIT */
    if (memFlashWidth == 2)
    {
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x0AAAA, 0x5555);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0x8080);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x0AAAA, 0x5555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x15554, 0x1010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
      }

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        result = BDI_GetWord(sectorAddr ^ 4L, &wValue);
        result = BDI_GetWord(sectorAddr, &wValue);
      } while ((wValue != 0xFFFF) && (result == BDI_OKAY));

      /* Reset chip */
      (void)BDI_SetWord(dwBaseAddr + 0x15554, 0xAAAA);
      (void)BDI_SetWord(dwBaseAddr + 0x0AAAA, 0x5555);
      (void)BDI_SetWord(dwBaseAddr + 0x15554, 0xF0F0);
    }
    break;


    case 4: /* 32BIT */
    if (memFlashWidth == 2)
    {
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x15554, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x15554, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x2AAA8, 0x10101010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      }

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        result = BDI_GetLong(sectorAddr ^ 4L, &lValue);
        result = BDI_GetLong(sectorAddr, &lValue);
      } while ((lValue != 0xFFFFFFFF) && (result == BDI_OKAY));

      /* Reset chip */
      (void)BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xAAAAAAAA);
      (void)BDI_SetLong(dwBaseAddr + 0x15554, 0x55555555);
      (void)BDI_SetLong(dwBaseAddr + 0x2AAA8, 0xF0F0F0F0);
    }
    else if (memFlashWidth == 4)
    {
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x10101010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      }

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        result = BDI_GetLong(sectorAddr ^ 4L, &lValue);
        result = BDI_GetLong(sectorAddr, &lValue);
      } while ((lValue != 0xFFFFFFFF) && (result == BDI_OKAY));

      /* Reset chip */
      (void)BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      (void)BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      (void)BDI_SetLong(dwBaseAddr + 0x015554, 0xF0F0F0F0);
    }
    break;

    default:
    result = BDI_ERR_INVALID_PARAMETER;
    break;
  } /* switch */

  return result;
} /* EraseAM29BDD */


/****************************************************************************
 ****************************************************************************

    Erase Spansion S29GL-S family

    INPUT:  mode            Erase mode
            sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

#define GLS_DEVICE_READY        BIT7
#define GLS_ERASE_STATUS        BIT5
#define GLS_PROG_STATUS         BIT4
#define GLS_SECTOR_LOCK         BIT1
#define GLS_ERASE_ERROR         (GLS_ERASE_STATUS | GLS_SECTOR_LOCK)
#define GLS_PROG_ERROR          (GLS_PROG_STATUS | GLS_SECTOR_LOCK)


static int EraseS29GLS(WORD mode, DWORD sectorAddr)
{
  int       result;
  WORD      wValue;
  DWORD     lValue;
  DWORD     cmdOffsetA;
  DWORD     cmdOffsetB;

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  result    = BDI_OKAY;
  startTime = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth) {
    case 2: /* 16 BIT */
    if (memFlashWidth == 2) {
      cmdOffsetA = 0x0AAA;
      cmdOffsetB = 0x0554;
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x0071);
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x00AA);
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetB, 0x0055);
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x0080);
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x00AA);
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetB, 0x0055);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x0010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x0030);
      }

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x0070);
        if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &wValue);
      } while (((wValue & GLS_DEVICE_READY) == 0) && (result == BDI_OKAY));

      /* check for erase error */
      if ((result == BDI_OKAY) && (wValue & GLS_ERASE_ERROR)) result = BDI_ERR_FLASH_ERASE;

      /* Reset chip */
      (void)BDI_SetWord(sectorAddr, 0x00F0);
    }
    break;


    case 4: /* 32BIT */
    if (memFlashWidth == 2) {
      cmdOffsetA = 0x1554;
      cmdOffsetB = 0x0AA8;
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00710071);
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00AA00AA);
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetB, 0x00550055);
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00800080);
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00AA00AA);
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetB, 0x00550055);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00100010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x00300030);
      }

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00700070);
        if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &lValue);
      } while (((lValue  & (lValue >> 16) & GLS_DEVICE_READY) == 0) && (result == BDI_OKAY));

      /* check for erase error */
      if ((result == BDI_OKAY) && ((lValue | (lValue >> 16)) & GLS_ERASE_ERROR)) result = BDI_ERR_FLASH_ERASE;

      /* Reset chip */
      (void)BDI_SetLong(sectorAddr, 0x00F000F0);
    }
    break;

    default:
    result = BDI_ERR_INVALID_PARAMETER;
    break;
  } /* switch */

  return result;
} /* EraseS29GLS */


/****************************************************************************
 ****************************************************************************

    Erase Spansion S29VS-R family

    INPUT:  mode            Erase mode
            sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

#define VSR_DEVICE_READY        BIT7
#define VSR_ERASE_STATUS        BIT5
#define VSR_PROG_STATUS         BIT4
#define VSR_SECTOR_LOCK         BIT1
#define VSR_BANK_STATUS         BIT0
#define VSR_ERASE_ERROR         (VSR_ERASE_STATUS | VSR_SECTOR_LOCK)
#define VSR_PROG_ERROR          (VSR_PROG_STATUS | VSR_SECTOR_LOCK)


static int EraseS29VSR(WORD mode, DWORD sectorAddr)
{
  int       result;
  WORD      wValue;
  DWORD     lValue;
  DWORD     cmdOffsetA;
  DWORD     cmdOffsetB;

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  result    = BDI_OKAY;
  startTime = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth) {
    case 2: /* 16 BIT */
    if (memFlashWidth == 2) {
      cmdOffsetA = 0x0AAA;
      cmdOffsetB = 0x0554;
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x0071);
      if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x0080);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetB, 0x0010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetB, 0x0030);
      }

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr + cmdOffsetA, 0x0070);
        if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &wValue);
      } while (((wValue & VSR_DEVICE_READY) == 0) && (result == BDI_OKAY));

      /* check for erase error */
      if ((result == BDI_OKAY) && (wValue & VSR_ERASE_ERROR)) result = BDI_ERR_FLASH_ERASE;

      /* Reset chip */
      (void)BDI_SetWord(sectorAddr, 0x00F0);
    }
    break;


    case 4: /* 32BIT */
    if (memFlashWidth == 2) {
      cmdOffsetA = 0x1554;
      cmdOffsetB = 0x0AA8;
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00710071);
      if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00800080);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetB, 0x00100010);
      }
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetB, 0x00300030);
      }

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr + cmdOffsetA, 0x00700070);
        if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &lValue);
      } while (((lValue  & (lValue >> 16) & VSR_DEVICE_READY) == 0) && (result == BDI_OKAY));

      /* check for erase error */
      if ((result == BDI_OKAY) && ((lValue | (lValue >> 16)) & VSR_ERASE_ERROR)) result = BDI_ERR_FLASH_ERASE;

      /* Reset chip */
      (void)BDI_SetLong(sectorAddr, 0x00F000F0);
    }
    break;

    default:
    result = BDI_ERR_INVALID_PARAMETER;
    break;
  } /* switch */

  return result;
} /* EraseS29VSR */


/****************************************************************************
 ****************************************************************************

    Erase AT49 family

    INPUT:  mode            Erase mode
            sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseAT49(WORD mode, DWORD sectorAddr)
{
  int       result;
  DWORD     dwBaseAddr;
  BYTE      newByte;
  WORD      newWord;
  DWORD     newLong;
  BYTE      newDL[8];

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  static const BYTE PatternA[]  = {0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA};
  static const BYTE PatternB[]  = {0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55};
  static const BYTE EraseCmd[]  = {0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80};
  static const BYTE ChipCmd[]   = {0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10};
  static const BYTE BlockCmd[]  = {0x50,0x50,0x50,0x50,0x50,0x50,0x50,0x50};
  static const BYTE SectorCmd[] = {0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30};
  static const BYTE ResetCmd[]  = {0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0};
  static const BYTE Erased[]    = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

  result     = BDI_OKAY;
  dwBaseAddr = sectorAddr & memBankMask;
  startTime  = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth)
  {
    case 1: /* 8 BIT */
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x002AAA, 0x55);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0x80);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x002AAA, 0x55);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0x10);
      } /* if */
      else if (mode == BDI_ERASE_BLOCK) {
        if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0x50);
      } /* else if */
      else {
        if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0x30);
      } /* else */

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_GetByte(sectorAddr, &newByte);
      } while ((result == BDI_OKAY) && (newByte != 0xFF));
      break;

    case 2: /* 16 BIT */
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x005554, 0x5555);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x8080);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xAAAA);
      if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x005554, 0x5555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x1010);
      } /* if */
      else if (mode == BDI_ERASE_BLOCK) {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x5050);
      } /* else if */
      else {
        if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
      } /* else */

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &newWord);
      } while ((result == BDI_OKAY) && (newWord != 0xFFFF));
      break;

    case 4: /* 32BIT */
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x80808080);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
      if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x10101010);
      } /* if */
      else if (mode == BDI_ERASE_BLOCK) {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x50505050);
      } /* else if */
      else {
        if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
      } /* else */

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &newLong);
      } while ((result == BDI_OKAY) && (newLong != 0xFFFFFFFF));
      break;

    case 8: /* 64BIT */
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, EraseCmd);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
      if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, ChipCmd);
      } /* if */
      else if (mode == BDI_ERASE_BLOCK) {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, BlockCmd);
      } /* else if */
      else {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
      } /* else */

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, newDL);
      } while ((result == BDI_OKAY) && (memcmp(newDL, Erased, 8) != 0));
      break;

  } /* switch */

  return result;
} /* EraseAT49 */


/****************************************************************************
 ****************************************************************************

    Erase AT49X8X16 family

    INPUT:  mode            Erase mode
            sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseAT49X8X16(WORD mode, DWORD sectorAddr)
{
  int       result;
  DWORD     dwBaseAddr;
  BYTE      newByte;
  WORD      newWord;
  DWORD     newLong;
  BYTE      newDL[8];

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  static const BYTE PatternA[]  = {0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA};
  static const BYTE PatternB[]  = {0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55};
  static const BYTE EraseCmd[]  = {0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80};
  static const BYTE ChipCmd[]   = {0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10};
  static const BYTE BlockCmd[]  = {0x50,0x50,0x50,0x50,0x50,0x50,0x50,0x50};
  static const BYTE SectorCmd[] = {0x30,0x30,0x30,0x30,0x30,0x30,0x30,0x30};
  static const BYTE ResetCmd[]  = {0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0};
  static const BYTE Erased[]    = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

  result     = BDI_OKAY;
  dwBaseAddr = sectorAddr & memBankMask;
  startTime  = times(&tbuf);
  if (mode == BDI_ERASE_CHIP) eraseTimeout *= 5;

  switch (memBusWidth)
  {
    case 1: /* 8 BIT */
      /* start erasing */
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x00AAAA, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0x55);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x00AAAA, 0x80);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x00AAAA, 0xAA);
      if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x005555, 0x55);
      if (mode == BDI_ERASE_CHIP) {
        if (result == BDI_OKAY) result = BDI_SetByte(dwBaseAddr + 0x00AAAA, 0x10);
      } /* if */
      else if (mode == BDI_ERASE_BLOCK) {
        if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0x50);
      } /* else if */
      else {
        if (result == BDI_OKAY) result = BDI_SetByte(sectorAddr, 0x30);
      } /* else */

      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_GetByte(sectorAddr, &newByte);
      } while ((result == BDI_OKAY) && (newByte != 0xFF));
      break;

    case 2: /* 16 BIT */
      if (memFlashWidth == 1)
      {
        /* start erasing */
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x015554, 0xAAAA);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x5555);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x015554, 0x8080);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x015554, 0xAAAA);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x5555);
        if (mode == BDI_ERASE_CHIP) {
          if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x015554, 0x1010);
        } /* if */
        else if (mode == BDI_ERASE_BLOCK) {
          if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x5050);
        } /* else if */
        else {
          if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
        } /* else */
      }
      else if (memFlashWidth == 2)
      {
        /* start erasing */
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xAAAA);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x005554, 0x5555);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x8080);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0xAAAA);
        if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x005554, 0x5555);
        if (mode == BDI_ERASE_CHIP) {
          if (result == BDI_OKAY) result = BDI_SetWord(dwBaseAddr + 0x00AAAA, 0x1010);
        } /* if */
        else if (mode == BDI_ERASE_BLOCK) {
          if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x5050);
        } /* else if */
        else {
          if (result == BDI_OKAY) result = BDI_SetWord(sectorAddr, 0x3030);
        } /* else */
      }
      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_GetWord(sectorAddr, &newWord);
      } while ((result == BDI_OKAY) && (newWord != 0xFFFF));
      break;

    case 4: /* 32BIT */
      if (memFlashWidth == 1)
      {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x02AAA8, 0xAAAAAAAA);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x55555555);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x02AAA8, 0x80808080);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x02AAA8, 0xAAAAAAAA);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x55555555);
        if (mode == BDI_ERASE_CHIP) {
          if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x02AAA8, 0x10101010);
        } /* if */
        else if (mode == BDI_ERASE_BLOCK) {
          if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x50505050);
        } /* else if */
        else {
          if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
        } /* else */
      }
      else if (memFlashWidth == 2)
      {
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x80808080);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0xAAAAAAAA);
        if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x00AAA8, 0x55555555);
        if (mode == BDI_ERASE_CHIP) {
          if (result == BDI_OKAY) result = BDI_SetLong(dwBaseAddr + 0x015554, 0x10101010);
        } /* if */
        else if (mode == BDI_ERASE_BLOCK) {
          if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x50505050);
        } /* else if */
        else {
          if (result == BDI_OKAY) result = BDI_SetLong(sectorAddr, 0x30303030);
        } /* else */
      }
      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_GetLong(sectorAddr, &newLong);
      } while ((result == BDI_OKAY) && (newLong != 0xFFFFFFFF));
      break;

    case 8: /* 64BIT */
      if (memFlashWidth == 1)
      {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, PatternA);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternB);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, EraseCmd);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, PatternA);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternB);
        if (mode == BDI_ERASE_CHIP) {
          if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x55550, BDI_AT_DEFAULT, 8, ChipCmd);
        } /* if */
        else if (mode == BDI_ERASE_BLOCK) {
          if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, BlockCmd);
        } /* else if */
        else {
          if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
        } /* else */
      }
      else if (memFlashWidth == 2)
      {
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, EraseCmd);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, PatternA);
        if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x15550, BDI_AT_DEFAULT, 8, PatternB);
        if (mode == BDI_ERASE_CHIP) {
          if (result == BDI_OKAY) result = BDI_MemoryWrite(dwBaseAddr + 0x2AAA8, BDI_AT_DEFAULT, 8, ChipCmd);
        } /* if */
        else if (mode == BDI_ERASE_BLOCK) {
          if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, BlockCmd);
        } /* else if */
        else {
          if (result == BDI_OKAY) result = BDI_MemoryWrite(sectorAddr, BDI_AT_DEFAULT, 8, SectorCmd);
        } /* else */
      }
      /* wait for end of erasing */
      do {
        if ((times(&tbuf) - startTime) > eraseTimeout) result = BDI_ERR_FLASH_ERASE;
        if (result == BDI_OKAY) result = BDI_MemoryRead(sectorAddr, BDI_AT_DEFAULT, 8, newDL);
      } while ((result == BDI_OKAY) && (memcmp(newDL, Erased, 8) != 0));
      break;
  } /* switch */

  return result;
} /* EraseAT49X8X16 */


/****************************************************************************
 ****************************************************************************

    Erase EraseSTR91xF internal flash

    INPUT:  mode      Erase mode
            addr      Startaddress of the sector, block or chip
    OUTPUT: return    0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseSTR910(WORD mode, DWORD addr)
{
  int           result;
  WORD          stateWord;

  struct tms    tbuf;
  clock_t       eraseTimeout = (ERASE_TIMEOUT * sysconf(_SC_CLK_TCK)) / 1000L;
  clock_t       startTime;

  result    = BDI_OKAY;
  startTime = times(&tbuf);

  /* start erasing */
  if (result == BDI_OKAY) result = BDI_SetWord(addr, (WORD)((mode == BDI_ERASE_BLOCK) ? 0x80 : 0x20));
  if (result == BDI_OKAY) result = BDI_SetWord(addr, 0xD0);

  /* wait for end of erasing */
  while (result == BDI_OKAY) {
    if (result == BDI_OKAY) result = BDI_GetWord(addr, &stateWord);
    if (result == BDI_OKAY) {
      /* check for finished */
      if ((stateWord & 0x80) == 0x80) {
        if ((stateWord & 0x20) != 0)
          result = BDI_ERR_FLASH_ERASE;
        else
          break;
      } /* if */
      /* check for timeout */
      else if ((times(&tbuf) - startTime) > eraseTimeout) {
        result = BDI_ERR_FLASH_ERASE;
      } /* else */
    } /* if */
  } /* while */

  /* Reset chip */
  BDI_SetWord(addr, 0x0050);
  BDI_SetWord(addr, 0xFFFF);

  return result;
} /* EraseSTR910 */


/****************************************************************************
 ****************************************************************************

    Erase Coldfire / M-Core / LPC2000 / MPC55xx flash module

    INPUT:  mode      Erase mode /Space select
            addr      Startaddress of the sector, block or chip
                      Sector Map for LPC2000 and MPC55xx
    OUTPUT: return    0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseFlashModule(WORD mode, DWORD addr)
{
  int   result;
  BYTE  eraseState;
  DWORD dwEraseAddr;

  /* start erasing */
  if (    (memFlashType != BDI_FLASH_H7F)
       && (memFlashType != BDI_FLASH_STM32F10)
       && (memFlashType != BDI_FLASH_LM3S)
       && (memFlashType != BDI_FLASH_STM32L15)
       && (memFlashType != BDI_FLASH_EFM32)
       && (memFlashType != BDI_FLASH_FTFL)
       && (memFlashType != BDI_FLASH_FTFE)
       && (memFlashType != BDI_FLASH_STM32F2)
       && (memFlashType != BDI_FLASH_STM32F4)
       && (memFlashType != BDI_FLASH_STM32W)
     ) {
    mode = (mode == BDI_ERASE_SECTOR) ? 1 : 0;
  } /* if */
  result = BDI_StartErase(memFlashType,
                          (BYTE)mode,
                          memFlashSize,
                          addr,
                          memRamBase,
                          memOption1);

  /* get erase result */
  eraseState = BDI_ERASE_ERASE;
  while ((result == BDI_OKAY) && (eraseState == BDI_ERASE_ERASE)) {
    result = BDI_GetEraseState(&eraseState, &dwEraseAddr);
  } /* if */

  /* analyse result */
  if (result == BDI_OKAY) {
    if (eraseState != BDI_ERASE_TERMINATED) {
      result = BDI_ERR_FLASH_ERASE;
    } /* if */
  } /* if */

  return result;
} /* EraseFlashModule */


/****************************************************************************
 ****************************************************************************

    Erase a flash sector

    INPUT:  sectorAddr      Startaddress of the sector to erase
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashEraseSector(DWORD sectorAddr)
{
  int   result;
  BOOL  mirrorV2;

  mirrorV2 = ((memFlashType == BDI_FLASH_S29M64X8) || (memFlashType == BDI_FLASH_S29M32X16));
  switch (memFlashType)
  {
    case BDI_FLASH_AM29F0X0:
    case BDI_FLASH_MIRROR:
      result = EraseAm29F010(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_I28FBXB:
    case BDI_FLASH_I28FBXW:
    case BDI_FLASH_I28FBXD:
    case BDI_FLASH_STRATAX8:
    case BDI_FLASH_STRATAX16:
      result = EraseI28FBX(sectorAddr);
      break;

    case BDI_FLASH_AM29FX00B:
    case BDI_FLASH_AM29FX00W:
    case BDI_FLASH_MIRRORX8:
    case BDI_FLASH_MIRRORX16:
    case BDI_FLASH_S29M64X8:
    case BDI_FLASH_S29M32X16:
      result = EraseAm29F100(mirrorV2, BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_AT49:
      result = EraseAT49(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_AT49X8:
    case BDI_FLASH_AT49X16:
      result = EraseAT49X8X16(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_AM29DX16:
    case BDI_FLASH_AM29DX32:
      result = EraseAM29D(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_AM29BDDX16:
    case BDI_FLASH_AM29BDDX32:
      result = EraseAM29BDD(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_S29GLSX16:
      result = EraseS29GLS(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_S29WSRX16:
    case BDI_FLASH_S29VSRX16:
      result = EraseS29VSR(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_MCORE_SGFM:
    case BDI_FLASH_CFM:
    case BDI_FLASH_CFM16:
    case BDI_FLASH_CFM32:
    case BDI_FLASH_LPC2000:
    case BDI_FLASH_STA2051:
    case BDI_FLASH_ADUC7000:
    case BDI_FLASH_ST30F:
    case BDI_FLASH_AT91SAM7S:
    case BDI_FLASH_STM32F10:
    case BDI_FLASH_LM3S:
    case BDI_FLASH_SAM9XE:
    case BDI_FLASH_SAM7L:
    case BDI_FLASH_STM32L15:
    case BDI_FLASH_EFM32:
    case BDI_FLASH_FTFL:
    case BDI_FLASH_FTFE:
    case BDI_FLASH_STM32F2:
    case BDI_FLASH_STM32F4:
    case BDI_FLASH_STM32W:
      result = EraseFlashModule(BDI_ERASE_SECTOR, sectorAddr);
      break;

    case BDI_FLASH_STR910F:
      result = EraseSTR910(BDI_ERASE_SECTOR, sectorAddr);
      break;

    default:
      result = BDI_ERR_INVALID_PARAMETER;
      break;
  } /* switch */

  return result;
} /* BDI_EraseSector */


/****************************************************************************
 ****************************************************************************

    Erase a flash sector, block or chip

    INPUT:  mode      Erase mode
            addr      Startaddress of the sector, block or chip
    OUTPUT: return    0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashErase(WORD mode, DWORD addr)
{
  int   result;
  BOOL  mirrorV2;

  mirrorV2 = ((memFlashType == BDI_FLASH_S29M64X8) || (memFlashType == BDI_FLASH_S29M32X16));
  switch (memFlashType)
  {
    case BDI_FLASH_AM29F0X0:
    case BDI_FLASH_MIRROR:
      result = EraseAm29F010(mode, addr);
      break;

    case BDI_FLASH_I28FBXB:
    case BDI_FLASH_I28FBXW:
    case BDI_FLASH_I28FBXD:
    case BDI_FLASH_STRATAX8:
    case BDI_FLASH_STRATAX16:
      result = EraseI28FBX(addr);
      break;

    case BDI_FLASH_AM29FX00B:
    case BDI_FLASH_AM29FX00W:
    case BDI_FLASH_MIRRORX8:
    case BDI_FLASH_MIRRORX16:
    case BDI_FLASH_S29M64X8:
    case BDI_FLASH_S29M32X16:
      result = EraseAm29F100(mirrorV2, mode, addr);
      break;

    case BDI_FLASH_AT49:
      result = EraseAT49(mode, addr);
      break;

    case BDI_FLASH_AT49X8:
    case BDI_FLASH_AT49X16:
      result = EraseAT49X8X16(mode, addr);
      break;

    case BDI_FLASH_AM29DX16:
    case BDI_FLASH_AM29DX32:
      result = EraseAM29D(mode, addr);
      break;

    case BDI_FLASH_AM29BDDX16:
    case BDI_FLASH_AM29BDDX32:
      result = EraseAM29BDD(mode, addr);
      break;

    case BDI_FLASH_S29GLSX16:
      result = EraseS29GLS(mode, addr);
      break;

    case BDI_FLASH_S29WSRX16:
    case BDI_FLASH_S29VSRX16:
      result = EraseS29VSR(mode, addr);
      break;

    case BDI_MCORE_SGFM:
    case BDI_FLASH_CFM:
    case BDI_FLASH_CFM16:
    case BDI_FLASH_CFM32:
    case BDI_FLASH_LPC2000:
    case BDI_FLASH_STA2051:
    case BDI_FLASH_ADUC7000:
    case BDI_FLASH_ST30F:
    case BDI_FLASH_AT91SAM7S:
    case BDI_FLASH_H7F:
    case BDI_FLASH_STM32F10:
    case BDI_FLASH_LM3S:
    case BDI_FLASH_SAM9XE:
    case BDI_FLASH_SAM7L:
    case BDI_FLASH_STM32L15:
    case BDI_FLASH_EFM32:
    case BDI_FLASH_FTFL:
    case BDI_FLASH_FTFE:
    case BDI_FLASH_STM32F2:
    case BDI_FLASH_STM32F4:
    case BDI_FLASH_STM32W:
      result = EraseFlashModule(mode, addr);
      break;

    case BDI_FLASH_STR910F:
      result = EraseSTR910(mode, addr);
      break;

    default:
      result = BDI_ERR_INVALID_PARAMETER;
      break;
  } /* switch */

  return result;
} /* BDI_FlashErase */


/****************************************************************************
 ****************************************************************************
                HC12 Flash/EEPROM Programming Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    Erase HC12 internal EEPROM or Flash memory

    INPUT:  flashAddr       the base address of the flash memory
            flashSize       the size of the flash memory
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

static int EraseIntFlashHC12(DWORD flashAddr, DWORD flashSize)
{
  int   result;
  BYTE  eraseState;
  DWORD dwEraseAddr;

  /* start erasing */
  result = BDI_StartErase(memFlashType,
                          1,
                          flashSize,
                          flashAddr,
                          memRamBase,
                          (DWORD)cpuTypeHC12);

  /* get erase result */
  if (result == BDI_OKAY) {
    result = BDI_GetEraseState(&eraseState, &dwEraseAddr);
  } /* if */

  /* analyse result */
  if (result == BDI_OKAY) {
    if (eraseState != BDI_ERASE_TERMINATED) {
      result = BDI_ERR_FLASH_ERASE;
    } /* if */
  } /* if */

  return result;
} /* EraseIntFlashHC12 */


/****************************************************************************
 ****************************************************************************

    Defines the  Flash memory for HC12 targets

    INPUT:  cpuType         the HC12 CPU type (see header file)
            flashType       the used Flash Memory Type
            memoryWidth     the memory bus width in bits (8,16,32)
            flashSize       the size of one flash memory chip in bytes
            workSpace       some programming algorithm must run in the target
                            CPU, enter the base address of a RAM area reseved
                            for this algorithm.
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashSetupHC12(WORD  cpuType,
                       WORD  flashType,
                       WORD  memoryWidth,
                       DWORD flashSize,
                       DWORD workSpace)
{
  int result;

  /* check parameter */
  pagedAddress = ((flashType & BDI_HC12_PAGED) != 0);
  flashType &= ~BDI_HC12_PAGED;
  if (    (flashType != BDI_FLASH_AM29F0X0)
       && (flashType != BDI_FLASH_I28FBXB)
       && (flashType != BDI_FLASH_I28FBXW)
       && (flashType != BDI_FLASH_AM29FX00B)
       && (flashType != BDI_FLASH_AM29FX00W)
       && (flashType != BDI_FLASH_AT49)
       && (flashType != BDI_FLASH_AT49X8)
       && (flashType != BDI_FLASH_AT49X16)
       && (flashType != BDI_FLASH_STRATAX8)
       && (flashType != BDI_FLASH_STRATAX16)
       && (flashType != BDI_HC12_EEPROM)
       && (flashType != BDI_HC12_FLASH)
       && (flashType != BDI_HC12A_EEPROM)
       && (flashType != BDI_HC12A_FLASH)
       && (flashType != BDI_STAR12_EEPROM)
       && (flashType != BDI_STAR12_FLASH)
       && (flashType != BDI_S12XE_PFLASH)
       && (flashType != BDI_S12XE_DFLASH)
     ) {
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* set bdiPro mode */
  result = BDI_SetProMode();
  if (result != BDI_OKAY) return result;

  /* setup flash parameters */
  targetIsHC12 = FALSE;
  targetIsS12X = FALSE;;
  if (   (cpuType == BDI_CPU_S12X_128K1)
      || (cpuType == BDI_CPU_S12X_256K2)
      || (cpuType == BDI_CPU_S12X_512K4)
      || (cpuType == BDI_CPU_S12XE)
      ) {
    targetIsS12X = TRUE;
  } /* if */
  else {
    targetIsHC12 = TRUE;
  } /* else */

  cpuTypeHC12  = cpuType;
  memOption1   = (DWORD)cpuType;
  memFlashType = (BYTE)flashType;
  memFlashSize = flashSize;
  flashChanged = TRUE;

  if (    (flashType == BDI_FLASH_I28FBXW)
       || (flashType == BDI_FLASH_AM29FX00W)
       || (flashType == BDI_FLASH_AT49X16)
       || (flashType == BDI_FLASH_STRATAX16)
       ) {
    memFlashWidth = 2;
  } /* if */
  else {
    memFlashWidth = 1;
  } /* else */

  if      (flashType == BDI_HC12_EEPROM)   memTypeHC12 = HC12_MEM_INT_EEPROM;
  else if (flashType == BDI_HC12_FLASH)    memTypeHC12 = HC12_MEM_INT_FLASH;
  else if (flashType == BDI_HC12A_EEPROM)  memTypeHC12 = HC12_MEM_INT_EEPROM;
  else if (flashType == BDI_HC12A_FLASH)   memTypeHC12 = HC12_MEM_INT_FLASH;
  else if (flashType == BDI_STAR12_EEPROM) memTypeHC12 = HC12_MEM_INT_EEPROM;
  else if (flashType == BDI_STAR12_FLASH)  memTypeHC12 = HC12_MEM_INT_FLASH;
  else if (flashType == BDI_S12XE_PFLASH)  memTypeHC12 = HC12_MEM_INT_FLASH;
  else if (flashType == BDI_S12XE_DFLASH)  memTypeHC12 = HC12_MEM_INT_FLASH;
  else                                     memTypeHC12 = HC12_MEM_EXTERNAL;

  if      (memoryWidth == 8)  memBusWidth = 1;
  else if (memoryWidth == 16) memBusWidth = 2;
  else if (memoryWidth == 32) memBusWidth = 4;
  else                        return BDI_ERR_INVALID_PARAMETER;

  memPageSize = 0;
  memRamBase  = workSpace;
  memBankMask = 0xFFFFFFFF - ( (flashSize * (DWORD)memBusWidth / (DWORD)memFlashWidth) - 1);

  return BDI_OKAY;
} /* BDI_FlashSetupHC12 */


/****************************************************************************
 ****************************************************************************

    Erase HC12 internal or external Flash memory

    INPUT:  flashAddr       the base address of the flash memory
            flashSize       the size of the flash memory
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashEraseHC12(DWORD flashAddr, DWORD flashSize)
{
  int result;

  /* Set read mode and memory type */
  memWriteMode = BDI_LOAD_MODE_RAM;
  result = BDI_SetWriteMode(BDI_LOAD_MODE_RAM, memTypeHC12, 0, 0, cpuTypeHC12);
  if (result != BDI_OKAY) return result;

  /* erase now */
  switch (memFlashType)
  {
    case BDI_FLASH_AM29F0X0:
      result = EraseAm29F010(BDI_ERASE_SECTOR, flashAddr);
      break;

    case BDI_FLASH_I28FBXB:
    case BDI_FLASH_I28FBXW:
    case BDI_FLASH_STRATAX8:
    case BDI_FLASH_STRATAX16:
      result = EraseI28FBX(flashAddr);
      break;

    case BDI_FLASH_AM29FX00B:
    case BDI_FLASH_AM29FX00W:
      result = EraseAm29F100(FALSE, BDI_ERASE_SECTOR, flashAddr);
      break;

    case BDI_FLASH_AT49:
      result = EraseAT49(BDI_ERASE_SECTOR, flashAddr);
      break;

    case BDI_FLASH_AT49X8:
    case BDI_FLASH_AT49X16:
      result = EraseAT49X8X16(BDI_ERASE_SECTOR, flashAddr);
      break;

    case BDI_HC12_EEPROM:
    case BDI_HC12_FLASH:
    case BDI_HC12A_EEPROM:
    case BDI_HC12A_FLASH:
    case BDI_STAR12_EEPROM:
    case BDI_STAR12_FLASH:
    case BDI_S12XE_PFLASH:
    case BDI_S12XE_DFLASH:
      result = EraseIntFlashHC12(flashAddr, flashSize);
      break;

    default:
      result = BDI_ERR_INVALID_PARAMETER;
      break;
  } /* switch */

  return result;
} /* BDI_FlashEraseHC12 */


/****************************************************************************
 ****************************************************************************
                MPC555 internal Flash Programming Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    Defines the programming parameters for MPC555 flash programming

    INPUT:  flashType       the used Flash Memory Type (16 or 17)
            cpuClock        the target CPU clock in hertz
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashSetup555(WORD  flashType,
                      DWORD cpuClock,
                      DWORD workSpace)
{
  int result;

  /* check parameter */
  if (    (flashType != BDI_MPC555_FLASH)
       && (flashType != BDI_MPC555_SHADOW)
     ) {
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* set bdiPro mode */
  result = BDI_SetProMode();
  if (result != BDI_OKAY) return result;

  /* setup flash parameters */
  memFlashType  = (BYTE)flashType;
  memBusWidth   = 2;
  memFlashWidth = 2;
  memPageSize   = 0;
  memBankMask   = 0;
  memRamBase    = workSpace;
  memOption1    = cpuClock;
  flashChanged  = TRUE;

  return BDI_OKAY;
} /* BDI_FlashSetup555 */


/****************************************************************************
 ****************************************************************************

    Erase MPC555 internal Flash memory

    The parameter blockMask has the following layout.

    +---------------+---------------+---------------+---------------+
    |   Module A    |   Module B    |               |               |
    +---------------+---------------+---------------+---------------+
    |0|1|2|3|4|5|6|7|0|1|2|3|4|5|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|
    +---------------+---------------+---------------+---------------+
    |31           24|23           16|15            8|7             0|

    If the appropriate bit is set, the block will be erased.

    INPUT:  flashBase       the base address of the first flash module
            blockMask       the mask with the flash block to erase
            clearCensor     if set the censor bits will be erased
            cpuClock        the target CPU clock in hertz
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashErase555(DWORD flashBase,
                      DWORD blockMask,
                      BOOL  clearCensor,
                      DWORD cpuClock,
                      DWORD workSpace)
{
  int   result;
  BYTE  eraseState;
  DWORD dwEraseAddr;

  /* start erasing */
  result = BDI_StartErase(BDI_MPC555_FLASH,
                          (BYTE)((clearCensor) ? 1 : 0),
                          blockMask,
                          flashBase,
                          workSpace,
                          cpuClock);

  /* get erase result */
  if (result == BDI_OKAY) {
    result = BDI_GetEraseState(&eraseState, &dwEraseAddr);
  } /* if */

  /* analyse result */
  if (result == BDI_OKAY) {
    if (eraseState != BDI_ERASE_TERMINATED) {
      result = BDI_ERR_FLASH_ERASE;
    } /* if */
  } /* if */

  return result;
} /* BDI_FlashErase555 */


/****************************************************************************
 ****************************************************************************

    Program MPC555 internal Flash memory

    INPUT:  fileName        the full path and filename of the S-record file
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

#define MPC555_FLASH_SIZE       0x70000

static BOOL DataToProgram(WORD count, const BYTE* pa)
{
  while (count--) {
    if (*pa != 0xFF) return TRUE;
    pa++;
  } /* while */
  return FALSE;
} /* DataToProgram */


int BDI_FlashWrite555(const char *fileName, DWORD *errorAddr)
{
  int   result;
  FILE* file;
  int   module;
  int   page;
  int   block;
  BYTE  blockMask;
  DWORD flashBase;
  DWORD moduleAddr;
  DWORD pageBase;
  DWORD pageAddr;
  BYTE  buffer[BDI_MAX_BLOCK_SIZE];
  BYTE* pData;
  BYTE* pImage;
  char  szLine[SREC_MAX_LINE];
  int   srecCount;
  DWORD srecAddr;
  BYTE  srecData[SREC_MAX_DATA];

  /* check correct setup */
  if (memFlashType != BDI_MPC555_FLASH) {
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* open S-Record file */
  file = fopen(fileName,"r");
  if (file == NULL) return BDI_ERR_FILE_ACCESS;

  /* allocate memory for image */
  pImage = malloc(MPC555_FLASH_SIZE);
  if (pImage == NULL) return BDI_ERR_MEM_ACCESS;
  (void)memset(pImage, 0xFF, MPC555_FLASH_SIZE);

  /* load image to local array */
  result = BDI_OKAY;
  flashBase = 0;
  while ((result == BDI_OKAY) && (fgets(szLine, sizeof szLine, file) != NULL)) {
    srecCount = DecodeSRecord(szLine, &srecAddr, srecData);
    if (srecCount > 0) {
      flashBase = srecAddr & 0xFFC00000;
      srecAddr -= flashBase;
      if ((srecAddr + (size_t)srecCount) <= MPC555_FLASH_SIZE) {
        (void)memcpy(pImage + srecAddr, srecData, (size_t)srecCount);
      } /* if */
      else {
        result = BDI_ERR_FILE_ACCESS;
      } /* else */
    } /* if */
    else if (srecCount < 0) {
      result = BDI_ERR_FILE_ACCESS;
    } /* else if */
  } /* while */
  fclose(file);
  if (result != BDI_OKAY) {
    free(pImage);
    return result;
  } /* if */

  /* set write mode */
  memWriteMode = memFlashType;
  result = BDI_SetWriteMode(BDI_MPC555_FLASH, 4, 0, memRamBase, memOption1 | 0x80000000);
  flashChanged = TRUE;

  /* init download */
  if (result == BDI_OKAY) result = BDI_InitDownload(BDI_NO_VERIFY, TRUE);
  if (result != BDI_OKAY) {
    free(pImage);
    return result;
  } /* if */

  /* process module A and B */
  moduleAddr = 0;
  for (module = 0; module < 2; module++) {
    pageBase = moduleAddr;
    for (page = 0; page < 512; page++) {
      blockMask = 0;
      pageAddr  = pageBase;
      pData     = buffer + 1;
      for (block = 0; block < 8; block++) {
        blockMask <<= 1;
        if (pageAddr >= MPC555_FLASH_SIZE) continue;
        if (DataToProgram(64, pImage + pageAddr)) {
          (void)memcpy(pData, pImage + pageAddr, 64);
          pData     += 64;
          blockMask |= 1;
        } /* else */
        pageAddr += 0x8000;
      } /* for */
      if (blockMask != 0) {
        buffer[0] = blockMask;
        result = BDI_Download(flashBase + pageBase, (WORD)(pData - buffer), buffer);
        if (result != BDI_OKAY) break;
      } /* if */
      pageBase += 64;
    } /* for */
    moduleAddr += 0x40000;
    if (result != BDI_OKAY) break;
  } /* for */

  /* terminate download */
  if ((result == BDI_OKAY) || (result == BDI_ERR_MEM_ACCESS)) {
    result = BDI_EndDownload(errorAddr);
  } /* if */

  free(pImage);
  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_FLASH_PROGRAM;
  return result;
} /* BDI_FlashWrite555 */


/****************************************************************************
 ****************************************************************************
                MPC56x internal Flash Programming Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    Defines the programming parameters for UC3F flash programming

    INPUT:  flashType       the used Flash Memory Type (25 or 26)
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashSetupUC3F(WORD flashType, DWORD workSpace)
{
  int result;

  /* check parameter */
  if (    (flashType != BDI_UC3F_FLASH)
       && (flashType != BDI_UC3F_SHADOW)
     ) {
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* set bdiPro mode */
  result = BDI_SetProMode();
  if (result != BDI_OKAY) return result;

  /* setup flash parameters */
  memFlashType  = (BYTE)flashType;
  memBusWidth   = 2;
  memFlashWidth = 2;
  memPageSize   = 0;
  memBankMask   = 0;
  memRamBase    = workSpace;
  memOption1    = 0;
  flashChanged  = TRUE;

  return BDI_OKAY;
} /* BDI_FlashSetupUC3F */


/****************************************************************************
 ****************************************************************************

    Erase MPC56x internal Flash memory

    The parameter blockMask has the following layout.

    +---------------+---------------+-----------+---+---------------+
    |               |               |           |SSB|   block[0:7]  |
    +---------------+---------------+-----------+---+---------------+
    |               |               |           |0 1|0 1 2 3 4 5 6 7|
    +---------------+---------------+-----------+---+---------------+
    |31           24|23           16|15            8|7             0|

    If the appropriate bit is set, the block will be erased.

    INPUT:  flashBase       the base address of the flash module
            blockMask       the mask with the flash block(s) to erase
            clearCensor     if set the censor bits will be erased
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashEraseUC3F(DWORD flashBase,
                       DWORD blockMask,
                       BOOL  clearCensor,
                       DWORD workSpace)
{
  int   result;
  BYTE  eraseState;
  DWORD dwEraseAddr;

  /* start erasing */
  result = BDI_StartErase(BDI_UC3F_FLASH,
                          (BYTE)((clearCensor) ? 1 : 0),
                          blockMask,
                          flashBase,
                          workSpace,
                          0);

  /* get erase result */
  if (result == BDI_OKAY) {
    result = BDI_GetEraseState(&eraseState, &dwEraseAddr);
  } /* if */

  /* analyse result */
  if (result == BDI_OKAY) {
    if (eraseState != BDI_ERASE_TERMINATED) {
      result = BDI_ERR_FLASH_ERASE;
    } /* if */
  } /* if */

  return result;
} /* BDI_FlashEraseUC3F */


/****************************************************************************
 ****************************************************************************
                MMC2107 internal Flash Programming Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    Defines the programming parameters for MPC555 flash programming

    INPUT:  flashType       the used Flash Memory Type (16 or 17)
            cpuClock        the target CPU clock in hertz
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashSetup2107(WORD  flashType,
                       DWORD cpuClock,
                       DWORD workSpace)
{
  int result;

  /* check parameter */
  if (    (flashType != BDI_CMFR_FLASH)
       && (flashType != BDI_CMFR_SHADOW)
     ) {
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* set bdiPro mode */
  result = BDI_SetProMode();
  if (result != BDI_OKAY) return result;

  /* setup flash parameters */
  memFlashType  = (BYTE)flashType;
  memBusWidth   = 4;
  memFlashWidth = 4;
  memPageSize   = 0;
  memBankMask   = 0;
  memRamBase    = workSpace;
  memOption1    = cpuClock;
  flashChanged  = TRUE;

  return BDI_OKAY;
} /* BDI_FlashSetup2107 */


/****************************************************************************
 ****************************************************************************

    Erase MMC2107 internal Flash memory

    The parameter blockMask has the following layout.

    +---------------+---------------+---------------+---------------+
    |               |               |               |  block[0:7]   |
    +---------------+---------------+---------------+---------------+
    |-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|-|0|1|2|3|4|5|6|7|
    +---------------+---------------+---------------+---------------+
    |31           24|23           16|15            8|7             0|

    If the appropriate bit is set, the block will be erased.

    INPUT:  flashBase       the base address of the first flash module
            blockMask       the mask with the flash block to erase
            cpuClock        the target CPU clock in hertz
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashErase2107(DWORD flashBase,
                       DWORD blockMask,
                       DWORD cpuClock,
                       DWORD workSpace)
{
  int   result;
  BYTE  eraseState;
  DWORD dwEraseAddr;

  /* start erasing */
  result = BDI_StartErase(BDI_CMFR_FLASH,
                          0,
                          blockMask,
                          flashBase,
                          workSpace,
                          cpuClock);

  /* get erase result */
  if (result == BDI_OKAY) {
    result = BDI_GetEraseState(&eraseState, &dwEraseAddr);
  } /* if */

  /* analyse result */
  if (result == BDI_OKAY) {
    if (eraseState != BDI_ERASE_TERMINATED) {
      result = BDI_ERR_FLASH_ERASE;
    } /* if */
  } /* if */

  return result;
} /* BDI_FlashErase2107 */


/****************************************************************************
 ****************************************************************************

    Program MMC2107 internal Flash memory

    INPUT:  fileName        the full path and filename of the S-record file
    OUTPUT: errorAddr       address of the failing byte
            return          0 if okay or a negativ number if error

 ****************************************************************************/

#define MMC2107_FLASH_SIZE       0x20000

int BDI_FlashWrite2107(const char *fileName, DWORD *errorAddr)
{
  int   result;
  FILE* file;
  int   page;
  int   block;
  BYTE  blockMask;
  DWORD flashBase;
  DWORD pageBase;
  DWORD pageAddr;
  BYTE  buffer[BDI_MAX_BLOCK_SIZE];
  BYTE* pData;
  BYTE* pImage;
  char  szLine[SREC_MAX_LINE];
  int   srecCount;
  DWORD srecAddr;
  BYTE  srecData[SREC_MAX_DATA];

  /* check correct setup */
  if (memFlashType != BDI_CMFR_FLASH) {
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* open S-Record file */
  file = fopen(fileName,"r");
  if (file == NULL) return BDI_ERR_FILE_ACCESS;

  /* allocate memory for image */
  pImage = malloc(MMC2107_FLASH_SIZE);
  if (pImage == NULL) return BDI_ERR_MEM_ACCESS;
  (void)memset(pImage, 0xFF, MMC2107_FLASH_SIZE);

  /* load image to local array */
  result = BDI_OKAY;
  flashBase = 0;
  while ((result == BDI_OKAY) && (fgets(szLine, sizeof szLine, file) != NULL)) {
    srecCount = DecodeSRecord(szLine, &srecAddr, srecData);
    if (srecCount > 0) {
      flashBase = srecAddr & 0xFFFE0000;
      srecAddr -= flashBase;
      if ((srecAddr + (size_t)srecCount) <= MMC2107_FLASH_SIZE) {
        (void)memcpy(pImage + srecAddr, srecData, (size_t)srecCount);
      } /* if */
      else {
        result = BDI_ERR_FILE_ACCESS;
      } /* else */
    } /* if */
    else if (srecCount < 0) {
      result = BDI_ERR_FILE_ACCESS;
    } /* else if */
  } /* while */
  fclose(file);
  if (result != BDI_OKAY) {
    free(pImage);
    return result;
  } /* if */

  /* set write mode */
  memWriteMode = memFlashType;
  result = BDI_SetWriteMode(BDI_CMFR_FLASH, 4, 0, memRamBase, memOption1 | 0x80000000);
  flashChanged = TRUE;

  /* init download */
  if (result == BDI_OKAY) result = BDI_InitDownload(BDI_NO_VERIFY, TRUE);
  if (result != BDI_OKAY) {
    free(pImage);
    return result;
  } /* if */

  /* process programming */
  pageBase = 0;
  for (page = 0; page < 256; page++) {
    blockMask = 0;
    pageAddr  = pageBase;
    pData     = buffer + 1;
    for (block = 0; block < 8; block++) {
      blockMask <<= 1;
      if (pageAddr >= MMC2107_FLASH_SIZE) continue;
      if (DataToProgram(64, pImage + pageAddr)) {
        (void)memcpy(pData, pImage + pageAddr, 64);
        pData     += 64;
        blockMask |= 1;
      } /* else */
      pageAddr += 0x4000;
    } /* for */
    if (blockMask != 0) {
      buffer[0] = blockMask;
      result = BDI_Download(flashBase + pageBase, (WORD)(pData - buffer), buffer);
      if (result != BDI_OKAY) break;
    } /* if */
    pageBase += 64;
  } /* for */

  /* terminate download */
  if ((result == BDI_OKAY) || (result == BDI_ERR_MEM_ACCESS)) {
    result = BDI_EndDownload(errorAddr);
  } /* if */

  free(pImage);
  if (result == BDI_ERR_MEM_ACCESS) result = BDI_ERR_FLASH_PROGRAM;
  return result;
} /* BDI_FlashWrite2107 */


/****************************************************************************
 ****************************************************************************
                MMC2114 internal Flash Programming Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    Defines the programming parameters for SGFM flash programming

    INPUT:  cpuClock        the target CPU clock in hertz
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashSetupSGFM(DWORD cpuClock, DWORD workSpace)
{
  int result;

  /* set bdiPro mode */
  result = BDI_SetProMode();
  if (result != BDI_OKAY) return result;

  /* setup flash parameters */
  memFlashType  = BDI_MCORE_SGFM;
  memBusWidth   = 4;
  memFlashWidth = 4;
  memPageSize   = 0;
  memBankMask   = 0;
  memRamBase    = workSpace;
  memOption1    = cpuClock;
  flashChanged  = TRUE;

  return BDI_OKAY;
} /* BDI_FlashSetupSGFM */


/****************************************************************************
 ****************************************************************************
                LPC2000 internal Flash Programming Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    Defines the programming parameters for LPC2000 flash programming

    INPUT:  flashSize       the flash size
            sysClock        the target system clock in kHz
            workSpace       address of workspace in target RAM
    OUTPUT: return          0 if okay or a negativ number if error

 ****************************************************************************/

int BDI_FlashSetupLPC2000(DWORD flashSize, DWORD sysClock, DWORD workSpace)
{

  /* setup flash parameters */
  memFlashType  = BDI_FLASH_LPC2000;
  memBusWidth   = 1;
  memFlashWidth = 1;
  memPageSize   = 0;
  memFlashSize  = flashSize;
  memBankMask   = flashSize;
  memRamBase    = workSpace;
  memOption1    = sysClock;
  flashChanged  = TRUE;

  return BDI_OKAY;
} /* BDI_FlashSetupLPC2000 */



